import abc
import copy

import torch
import torch.nn as nn
import torch.nn.functional as F

from scipy.stats import ttest_rel
from tensordict import TensorDict
from torch.utils.data import DataLoader, Dataset

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.common.critic import CriticNetwork, create_critic_from_actor
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class REINFORCEBaseline(nn.Module, metaclass=abc.ABCMeta):
    """Base class for REINFORCE baselines"""

    def __init__(self, *args, **kw):
        super().__init__()
        pass

    def wrap_dataset(self, dataset: Dataset, *args, **kw):
        """Wrap dataset with baseline-specific functionality"""
        return dataset

    @abc.abstractmethod
    def eval(
        self, td: TensorDict, reward: torch.Tensor, env: RL4COEnvBase = None, **kwargs
    ):
        """Evaluate baseline"""
        raise NotImplementedError

    def epoch_callback(self, *args, **kw):
        """Callback at the end of each epoch
        For example, update baseline parameters and obtain baseline values
        """
        pass

    def setup(self, *args, **kw):
        """To be called before training during setup phase
        This follow PyTorch Lightning's setup() convention
        """
        pass


class NoBaseline(REINFORCEBaseline):
    """No baseline: return 0 for baseline and neg_los"""

    def eval(self, td, reward, env=None):
        return 0, 0  # No baseline, no neg_los


class SharedBaseline(REINFORCEBaseline):
    """Shared baseline: return mean of reward as baseline"""

    def eval(self, td, reward, env=None, on_dim=1):  # e.g. [batch, pomo, ...]
        return reward.mean(dim=on_dim, keepdims=True), 0


class ExponentialBaseline(REINFORCEBaseline):
    """Exponential baseline: return exponential moving average of reward as baseline

    Args:
        beta: Beta value for the exponential moving average
    """

    def __init__(self, beta=0.8, **kw):
        super(REINFORCEBaseline, self).__init__()

        self.beta = beta
        self.v = None

    def eval(self, td, reward, env=None):
        if self.v is None:
            v = reward.mean()
        else:
            v = self.beta * self.v + (1.0 - self.beta) * reward.mean()
        self.v = v.detach()  # Detach since we never want to backprop
        return self.v, 0  # No loss


class MeanBaseline(REINFORCEBaseline):
    """Mean baseline: return mean of reward as baseline"""

    def __new__(cls, **kw):
        return ExponentialBaseline(beta=0.0, **kw)


class WarmupBaseline(REINFORCEBaseline):
    """Warmup baseline: return convex combination of baseline and exponential baseline

    Args:
        baseline: Baseline to use after warmup
        n_epochs: Number of epochs to warmup
        warmup_exp_beta: Beta value for the exponential baseline during warmup
    """

    def __init__(self, baseline, n_epochs=1, warmup_exp_beta=0.8, **kw):
        super(REINFORCEBaseline, self).__init__()

        self.baseline = baseline
        assert n_epochs > 0, "n_epochs to warmup must be positive"
        self.warmup_baseline = ExponentialBaseline(warmup_exp_beta)
        self.alpha = 0
        self.n_epochs = n_epochs

    def wrap_dataset(self, dataset, *args, **kw):
        if self.alpha > 0:
            return self.baseline.wrap_dataset(dataset, *args, **kw)
        return self.warmup_baseline.wrap_dataset(dataset, *args, **kw)

    def setup(self, *args, **kw):
        self.baseline.setup(*args, **kw)

    def eval(self, td, reward, env=None):
        if self.alpha == 1:
            return self.baseline.eval(td, reward, env)
        if self.alpha == 0:
            return self.warmup_baseline.eval(td, reward, env)
        v_b, l_b = self.baseline.eval(td, reward, env)
        v_wb, l_wb = self.warmup_baseline.eval(td, reward, env)
        # Return convex combination of baseline and of loss
        return (
            self.alpha * v_b + (1 - self.alpha) * v_wb,
            self.alpha * l_b + (1 - self.alpha) * l_wb,
        )

    def epoch_callback(self, *args, **kw):
        # Need to call epoch callback of inner policy (also after first epoch if we have not used it)
        self.baseline.epoch_callback(*args, **kw)
        if kw["epoch"] < self.n_epochs:
            self.alpha = (kw["epoch"] + 1) / float(self.n_epochs)
            log.info("Set warmup alpha = {}".format(self.alpha))


class CriticBaseline(REINFORCEBaseline):
    """Critic baseline: use critic network as baseline

    Args:
        critic: Critic network to use as baseline. If None, create a new critic network based on the environment
    """

    def __init__(self, critic: CriticNetwork = None, **unused_kw):
        super(CriticBaseline, self).__init__()
        self.critic = critic

    def setup(self, policy, env, **kwargs):
        if self.critic is None:
            log.info("Critic not found. Creating critic network for {}".format(env.name))
            self.critic = create_critic_from_actor(policy)

    def eval(self, x, c, env=None):
        v = self.critic(x).squeeze(-1)
        # detach v since actor should not backprop through baseline, only for loss
        return v.detach(), F.mse_loss(v, c.detach())


class RolloutBaseline(REINFORCEBaseline):
    """Rollout baseline: use greedy rollout as baseline

    Args:
        bl_alpha: Alpha value for the baseline T-test
    """

    def __init__(self, bl_alpha=0.05, **kw):
        super(RolloutBaseline, self).__init__()
        self.bl_alpha = bl_alpha

    def setup(self, *args, **kw):
        self._update_policy(*args, **kw)

    def _update_policy(
        self, policy, env, batch_size=64, device="cpu", dataset_size=None, dataset=None
    ):
        """Update policy (=actor) and rollout baseline values"""
        self.policy = copy.deepcopy(policy).to(device)
        if dataset is None:
            log.info("Creating evaluation dataset for rollout baseline")
            self.dataset = env.dataset(batch_size=[dataset_size])

        log.info("Evaluating baseline policy on evaluation dataset")
        self.bl_vals = (
            self.rollout(self.policy, env, batch_size, device, self.dataset).cpu().numpy()
        )
        self.mean = self.bl_vals.mean()

    def eval(self, td, reward, env):
        """Evaluate rollout baseline

        Warning:
            This is not differentiable and should only be used for evaluation.
            Also, it is recommended to use the `rollout` method directly instead of this method.
        """
        with torch.inference_mode():
            reward = self.policy(td, env)["reward"]
        return reward, 0

    def epoch_callback(
        self, policy, env, batch_size=64, device="cpu", epoch=None, dataset_size=None
    ):
        """Challenges the current baseline with the policy and replaces the baseline policy if it is improved"""
        log.info("Evaluating candidate policy on evaluation dataset")
        candidate_vals = self.rollout(policy, env, batch_size, device).cpu().numpy()
        candidate_mean = candidate_vals.mean()

        log.info(
            "Candidate mean: {:.3f}, Baseline mean: {:.3f}".format(
                candidate_mean, self.mean
            )
        )
        if candidate_mean - self.mean > 0:
            # Calc p value with inverse logic (costs)
            t, p = ttest_rel(-candidate_vals, -self.bl_vals)

            p_val = p / 2  # one-sided
            assert t < 0, "T-statistic should be negative"
            log.info("p-value: {:.3f}".format(p_val))
            if p_val < self.bl_alpha:
                log.info("Updating baseline")
                self._update_policy(policy, env, batch_size, device, dataset_size)

    def rollout(self, policy, env, batch_size=64, device="cpu", dataset=None):
        """Rollout the policy on the given dataset"""

        # if dataset is None, use the dataset of the baseline
        dataset = self.dataset if dataset is None else dataset

        policy.eval()
        policy = policy.to(device)

        def eval_policy(batch):
            with torch.inference_mode():
                batch = env.reset(batch.to(device))
                return policy(batch, env, decode_type="greedy")["reward"]

        dl = DataLoader(dataset, batch_size=batch_size, collate_fn=dataset.collate_fn)

        rewards = torch.cat([eval_policy(batch) for batch in dl], 0)
        return rewards

    def wrap_dataset(self, dataset, env, batch_size=64, device="cpu", **kw):
        """Wrap the dataset in a baseline dataset

        Note:
            This is an alternative to `eval` that does not require the policy to be passed
            at every call but just once. Values are added to the dataset. This also allows for
            larger batch sizes since we evauate the policy without gradients.
        """
        rewards = (
            self.rollout(self.policy, env, batch_size, device, dataset=dataset)
            .detach()
            .cpu()
        )
        return dataset.add_key("extra", rewards)

    def __getstate__(self):
        """Do not include datasets in state to avoid pickling issues"""
        state = self.__dict__.copy()
        try:
            del state["dataset"]
        except KeyError:
            pass
        return state

    def __setstate__(self, state):
        """Restore datasets after unpickling. Will be restored in setup"""
        self.__dict__.update(state)
        self.dataset = None


REINFORCE_BASELINES_REGISTRY = {
    "no": NoBaseline,
    "shared": SharedBaseline,
    "exponential": ExponentialBaseline,
    "critic": CriticBaseline,
    "mean": MeanBaseline,
    "rollout_only": RolloutBaseline,
    "warmup": WarmupBaseline,
}


def get_reinforce_baseline(name, **kw):
    """Get a REINFORCE baseline by name
    The rollout baseline default to warmup baseline with one epoch of
    exponential baseline and the greedy rollout
    """
    if name == "warmup":
        inner_baseline = kw.get("baseline", "rollout")
        if not isinstance(inner_baseline, REINFORCEBaseline):
            inner_baseline = get_reinforce_baseline(inner_baseline, **kw)
        return WarmupBaseline(inner_baseline, **kw)
    elif name == "rollout":
        warmup_epochs = kw.get("n_epochs", 1)
        warmup_exp_beta = kw.get("exp_beta", 0.8)
        bl_alpha = kw.get("bl_alpha", 0.05)
        return WarmupBaseline(
            RolloutBaseline(bl_alpha=bl_alpha), warmup_epochs, warmup_exp_beta
        )

    if name is None:
        name = "no"  # default to no baseline
    baseline_cls = REINFORCE_BASELINES_REGISTRY.get(name, None)
    if baseline_cls is None:
        raise ValueError(
            f"Unknown baseline {baseline_cls}. Available baselines: {REINFORCE_BASELINES_REGISTRY.keys()}"
        )
    return baseline_cls(**kw)



import time

from functools import partial
from typing import Any

import torch

from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

from rl4co.data.transforms import StateAugmentation
from rl4co.models.common.transductive import TransductiveModel
from rl4co.models.zoo.eas.decoder import forward_eas, forward_pointer_attn_eas_lay
from rl4co.models.zoo.eas.nn import EASLayerNet
from rl4co.utils.decoding import get_log_likelihood
from rl4co.utils.ops import batchify, gather_by_index, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class EAS(TransductiveModel):
    """Efficient Active Search for Neural Combination Optimization from Hottung et al. (2022).
    Fine-tunes a subset of parameters (such as node embeddings or newly added layers) thus avoiding
    expensive re-encoding of the problem.
    Reference: https://openreview.net/pdf?id=nO5caZwFwYu

    Args:
        env: RL4CO environment to be solved
        policy: policy network
        dataset: dataset to be used for training
        use_eas_embedding: whether to use EAS embedding (EASEmb)
        use_eas_layer: whether to use EAS layer (EASLay)
        eas_emb_cache_keys: keys to cache in the embedding
        eas_lambda: lambda parameter for IL loss
        batch_size: batch size for training
        max_iters: maximum number of iterations
        augment_size: number of augmentations per state
        augment_dihedral: whether to augment with dihedral rotations
        parallel_runs: number of parallel runs
        baseline: REINFORCE baseline type (multistart, symmetric, full)
        max_runtime: maximum runtime in seconds
        save_path: path to save solution checkpoints
        optimizer: optimizer to use for training
        optimizer_kwargs: keyword arguments for optimizer
        verbose: whether to print progress for each iteration
    """

    def __init__(
        self,
        env,
        policy,
        dataset: Dataset | str,
        use_eas_embedding: bool = True,
        use_eas_layer: bool = False,
        eas_emb_cache_keys: list[str] = ["logit_key"],
        eas_lambda: float = 0.013,
        batch_size: int = 2,
        max_iters: int = 200,
        augment_size: int = 8,
        augment_dihedral: bool = True,
        num_parallel_runs: int = 1,
        baseline: str = "multistart",
        max_runtime: int = 86_400,
        save_path: str = None,
        optimizer: str | torch.optim.Optimizer | partial = "Adam",
        optimizer_kwargs: dict = {"lr": 0.0041, "weight_decay": 1e-6},
        verbose: bool = True,
        **kwargs,
    ):
        self.save_hyperparameters(logger=False, ignore=["env", "policy", "dataset"])

        assert (
            self.hparams.use_eas_embedding or self.hparams.use_eas_layer
        ), "At least one of `use_eas_embedding` or `use_eas_layer` must be True."

        super(EAS, self).__init__(
            env,
            policy=policy,
            dataset=dataset,
            batch_size=batch_size,
            max_iters=max_iters,
            max_runtime=max_runtime,
            save_path=save_path,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            **kwargs,
        )

        assert self.hparams.baseline in [
            "multistart",
            "symmetric",
            "full",
        ], f"Baseline {self.hparams.baseline} not supported."

    def setup(self, stage="fit"):
        """Setup base class and instantiate:
        - augmentation
        - instance solutions and rewards
        - original policy state dict
        """
        log.info(
            f"Setting up Efficient Active Search (EAS) with: \n"
            f"- EAS Embedding: {self.hparams.use_eas_embedding} \n"
            f"- EAS Layer: {self.hparams.use_eas_layer} \n"
        )
        super(EAS, self).setup(stage)

        # Instantiate augmentation
        self.augmentation = StateAugmentation(
            num_augment=self.hparams.augment_size,
            augment_fn="dihedral8" if self.hparams.augment_dihedral else "symmetric",
        )

        # Store original policy state dict
        self.original_policy_state = self.policy.state_dict()

        # Get dataset size and problem size
        len(self.dataset)
        _batch = next(iter(self.train_dataloader()))
        self.problem_size = self.env.reset(_batch)["action_mask"].shape[-1]
        self.instance_solutions = []
        self.instance_rewards = []

    def on_train_batch_start(self, batch: Any, batch_idx: int):
        """Called before training (i.e. search) for a new batch begins.
        We re-load the original policy state dict and configure all parameters not to require gradients.
        We do the rest in the training step.
        """
        self.policy.load_state_dict(self.original_policy_state)

        # Set all policy parameters to not require gradients
        for param in self.policy.parameters():
            param.requires_grad = False

    def training_step(self, batch, batch_idx):
        """Main search loop. We use the training step to effectively adapt to a `batch` of instances."""
        # Augment state
        batch_size = batch.shape[0]
        td_init = self.env.reset(batch)
        n_aug, n_start, n_runs = (
            self.augmentation.num_augment,
            self.env.get_num_starts(td_init),
            self.hparams.num_parallel_runs,
        )
        td_init = self.augmentation(td_init)
        td_init = batchify(td_init, n_runs)
        num_instances = batch_size * n_aug * n_runs  # NOTE: no num_starts!
        # batch_r = n_runs * batch_size # effective batch size
        group_s = (
            n_start + 1
        )  # number of different rollouts per instance (+1 for incumbent solution construction)

        # Get encoder and decoder for simplicity
        encoder = self.policy.encoder
        decoder = self.policy.decoder

        # Precompute the cache of the embeddings (i.e. q,k,v and logit_key)
        embeddings, _ = encoder(td_init)
        cached_embeds = decoder._precompute_cache(embeddings)

        # Collect optimizer parameters
        opt_params = []
        if self.hparams.use_eas_layer:
            # EASLay: replace forward of logit attention computation. EASLayer
            eas_layer = EASLayerNet(num_instances, decoder.embed_dim).to(batch.device)
            decoder.pointer.eas_layer = partial(eas_layer, decoder.pointer)
            decoder.pointer.forward = partial(
                forward_pointer_attn_eas_lay, decoder.pointer
            )
            for param in eas_layer.parameters():
                opt_params.append(param)
        if self.hparams.use_eas_embedding:
            # EASEmb: set gradient of emb_key to True
            # for all the keys, wrap the embedding in a nn.Parameter
            for key in self.hparams.eas_emb_cache_keys:
                setattr(
                    cached_embeds, key, torch.nn.Parameter(getattr(cached_embeds, key))
                )
                opt_params.append(getattr(cached_embeds, key))
        decoder.forward_eas = partial(forward_eas, decoder)

        # We pass attributes saved in policy too
        def set_attr_if_exists(attr):
            if hasattr(self.policy, attr):
                setattr(decoder, attr, getattr(self.policy, attr))

        for attr in ["temperature", "tanh_clipping", "mask_logits"]:
            set_attr_if_exists(attr)

        self.configure_optimizers(opt_params)

        # Solution and reward buffer
        max_reward = torch.full((batch_size,), -float("inf"), device=batch.device)
        best_solutions = torch.zeros(
            batch_size, self.problem_size * 2, device=batch.device, dtype=int
        )  # i.e. incumbent solutions

        # Init search
        t_start = time.time()
        for iter_count in range(self.hparams.max_iters):
            # Evaluate policy with sampling multistarts passing the cached embeddings
            best_solutions_expanded = best_solutions.repeat(n_aug, 1).repeat(n_runs, 1)
            logprobs, actions, td_out, reward = decoder.forward_eas(
                td_init.clone(),
                cached_embeds=cached_embeds,
                best_solutions=best_solutions_expanded,
                iter_count=iter_count,
                env=self.env,
                decode_type="multistart_sampling",
                num_starts=n_start,
            )

            # Unbatchify to get correct dimensions
            ll = get_log_likelihood(logprobs, actions, td_out.get("mask", None))
            ll = unbatchify(ll, (n_runs * batch_size, n_aug, group_s)).squeeze()
            reward = unbatchify(reward, (n_runs * batch_size, n_aug, group_s)).squeeze()
            actions = unbatchify(actions, (n_runs * batch_size, n_aug, group_s)).squeeze()

            # Compute REINFORCE loss with shared baselines
            # compared to original EAS, we also support symmetric and full baselines
            group_reward = reward[..., :-1]  # exclude incumbent solution
            if self.hparams.baseline == "multistart":
                bl_val = group_reward.mean(dim=-1, keepdim=True)
            elif self.hparams.baseline == "symmetric":
                bl_val = group_reward.mean(dim=-2, keepdim=True)
            elif self.hparams.baseline == "full":
                bl_val = group_reward.mean(dim=-1, keepdim=True).mean(
                    dim=-2, keepdim=True
                )
            else:
                raise ValueError(f"Baseline {self.hparams.baseline} not supported.")

            # REINFORCE loss
            advantage = group_reward - bl_val
            loss_rl = -(advantage * ll[..., :-1]).mean()
            # IL loss
            loss_il = -ll[..., -1].mean()
            # Total loss
            loss = loss_rl + self.hparams.eas_lambda * loss_il

            # Manual backpropagation
            opt = self.optimizers()
            opt.zero_grad()
            self.manual_backward(loss)

            # Save best solutions and rewards
            # Get max reward for each group and instance
            max_reward = reward.max(dim=2)[0].max(dim=1)[0]

            # Reshape and rank rewards
            reward_group = reward.reshape(n_runs * batch_size, -1)
            _, top_indices = torch.topk(reward_group, k=1, dim=1)

            # Obtain best solutions found so far
            solutions = actions.reshape(n_runs * batch_size, n_aug * group_s, -1)
            best_solutions_iter = gather_by_index(solutions, top_indices, dim=1)
            best_solutions[:, : best_solutions_iter.shape[1]] = best_solutions_iter

            self.log_dict(
                {
                    "loss": loss,
                    "max_reward": max_reward.mean(),
                    "step": iter_count,
                    "time": time.time() - t_start,
                },
                on_step=self.log_on_step,
            )

            log.info(
                f"{iter_count}/{self.hparams.max_iters} | "
                f" Reward: {max_reward.mean().item():.2f} "
            )

            # Stop if max runtime is exceeded
            if time.time() - t_start > self.hparams.max_runtime:
                log.info(f"Max runtime of {self.hparams.max_runtime} seconds exceeded.")
                break

        return {"max_reward": max_reward, "best_solutions": best_solutions}

    def on_train_batch_end(
        self, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        """We store the best solution and reward found."""
        max_rewards, best_solutions = outputs["max_reward"], outputs["best_solutions"]
        self.instance_solutions.append(best_solutions)
        self.instance_rewards.append(max_rewards)
        log.info(f"Best reward: {max_rewards.mean():.2f}")

    def on_train_epoch_end(self) -> None:
        """Called when the train ends."""
        save_path = self.hparams.save_path
        # concatenate solutions and rewards
        self.instance_solutions = pad_sequence(
            self.instance_solutions, batch_first=True, padding_value=0
        ).squeeze()
        self.instance_rewards = torch.cat(self.instance_rewards, dim=0).squeeze()
        if save_path is not None:
            log.info(f"Saving solutions and rewards to {save_path}...")
            torch.save(
                {"solutions": self.instance_solutions, "rewards": self.instance_rewards},
                save_path,
            )

        # https://github.com/Lightning-AI/lightning/issues/1406
        self.trainer.should_stop = True


class EASEmb(EAS):
    """EAS with embedding adaptation"""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if not kwargs.get("use_eas_embedding", False) or kwargs.get(
            "use_eas_layer", True
        ):
            log.warning(
                "Setting `use_eas_embedding` to True and `use_eas_layer` to False. Use EAS base class to override."
            )
        kwargs["use_eas_embedding"] = True
        kwargs["use_eas_layer"] = False
        super(EASEmb, self).__init__(*args, **kwargs)


class EASLay(EAS):
    """EAS with layer adaptation"""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if kwargs.get("use_eas_embedding", False) or not kwargs.get(
            "use_eas_layer", True
        ):
            log.warning(
                "Setting `use_eas_embedding` to True and `use_eas_layer` to False. Use EAS base class to override."
            )
        kwargs["use_eas_embedding"] = False
        kwargs["use_eas_layer"] = True
        super(EASLay, self).__init__(*args, **kwargs)



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_distance
from rl4co.utils.pylogger import get_pylogger

from .generator import SVRPGenerator

log = get_pylogger(__name__)


class SVRPEnv(RL4COEnvBase):
    """Skill-Vehicle Routing Problem (SVRP) environment.
    Basic Skill-VRP environment. The environment is a variant of the Capacitated Vehicle Routing Problem (CVRP).
    Each technician has a certain skill-level and each customer node requires a certain skill-level to be serviced.
    Each customer node needs is to be serviced by exactly one technician. Technicians can only service nodes if
    their skill-level is greater or equal to the required skill-level of the node. The environment is episodic and
    the goal is to minimize the total travel cost of the technicians. The travel cost depends on the skill-level of
    the technician. The environment is defined by the following parameters:

    Observations:
        - locations of the depot, pickup, and delivery locations
        - current location of the vehicle
        - the remaining locations to deliver
        - the visited locations
        - the current step

    Constraints:
        - the tour starts and ends at the depot
        - each pickup location must be visited before its corresponding delivery location
        - the vehicle cannot visit the same location twice

    Finish Condition:
        - the vehicle has visited all locations

    Reward:
        - (minus) the negative length of the path

    Args:
        generator: PDPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "svrp"

    def __init__(
        self,
        generator: SVRPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = SVRPGenerator(**generator_params)
        self.generator = generator
        self.tech_costs = self.generator.tech_costs
        self._make_spec(self.generator)

    def _make_spec(self, generator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            skills=Bounded(
                low=generator.min_skill,
                high=generator.max_skill,
                shape=(generator.num_loc, 1),
                dtype=torch.float32,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc + 1, 1),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc + 1,
        )
        self.reward_spec = Unbounded(shape=(1,), dtype=torch.float32)
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """Calculates the action mask for the Skill-VRP. The action mask is a binary mask that indicates which customer nodes can be services, given the previous decisions.
        For the Skill-VRP, a node can be serviced if the technician has the required skill-level and the node has not been visited yet.
        The depot cannot be visited if there are still unserved nodes and the technician either just visited the depot or is the last technician
        (because every customer node needs to be visited).
        """
        batch_size = td["locs"].shape[0]
        # check skill level
        current_tech_skill = gather_by_index(td["techs"], td["current_tech"]).reshape(
            [batch_size, 1]
        )
        can_service = td["skills"] <= current_tech_skill.unsqueeze(1).expand_as(
            td["skills"]
        )
        mask_loc = td["visited"][..., 1:, :].to(can_service.dtype) | ~can_service
        # Cannot visit the depot if there are still unserved nodes and I either just visited the depot or am the last technician
        mask_depot = (
            (td["current_node"] == 0) | (td["current_tech"] == td["techs"].size(-2) - 1)
        ) & ((mask_loc == 0).int().sum(-2) > 0)
        return ~torch.cat((mask_depot[..., None], mask_loc), -2).squeeze(-1)

    def _step(self, td: TensorDict) -> torch.Tensor:
        """Step function for the Skill-VRP. If a technician returns to the depot, the next technician is sent out.
        The visited node is marked as visited. The reward is set to zero and the done flag is set if all nodes have been visited.
        """
        current_node = td["action"][:, None]  # Add dimension for step

        # if I go back to the depot, send out next technician
        td["current_tech"] += (current_node == 0).int()

        # Add one dimension since we write a single value
        visited = td["visited"].scatter(-2, current_node[..., None], 1)

        # SECTION: get done
        done = visited.sum(-2) == visited.size(-2)
        reward = torch.zeros_like(done)

        td.update(
            {
                "current_node": current_node,
                "visited": visited,
                "reward": reward,
                "done": done,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self, td: Optional[TensorDict] = None, batch_size: Optional[list] = None
    ) -> TensorDict:
        device = td.device

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": torch.cat((td["depot"][:, None, :], td["locs"]), -2),
                "techs": td["techs"],
                "skills": td["skills"],
                "current_node": torch.zeros(
                    *batch_size, 1, dtype=torch.long, device=device
                ),
                "current_tech": torch.zeros(
                    *batch_size, 1, dtype=torch.long, device=device
                ),
                "visited": torch.zeros(
                    (*batch_size, td["locs"].shape[-2] + 1, 1),
                    dtype=torch.uint8,
                    device=device,
                ),
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        """Calculated the reward, where the reward is the negative total travel cost of the technicians.
        The travel cost depends on the skill-level of the technician."""
        # Check that the solution is valid
        if self.check_solution:
            self.check_solution_validity(td, actions)

        # Gather dataset in order of tour
        batch_size = td["locs"].shape[0]
        depot = td["locs"][..., 0:1, :]
        locs_ordered = torch.cat(
            [
                depot,
                gather_by_index(td["locs"], actions).reshape(
                    [batch_size, actions.size(-1), 2]
                ),
            ],
            dim=1,
        )

        # calculate travelling costs depending on the technicians' skill level
        costs = torch.zeros(batch_size, locs_ordered.size(-2), device=self.device)
        indices = torch.nonzero(actions == 0)
        start = tech = 0
        batch = 0
        for each in indices:
            if each[0] > batch:
                costs[batch, start:] = self.tech_costs[tech]
                start = tech = 0
                batch = each[0]
            end = (
                each[-1] + 1
            )  # indices in locs_ordered are shifted by one due to added depot in the front
            costs[batch, start:end] = self.tech_costs[tech]
            tech += 1
            start = end
        costs[batch, start:] = self.tech_costs[tech]

        travel_to = torch.roll(locs_ordered, -1, dims=-2)
        distances = get_distance(locs_ordered, travel_to)
        return -(distances * costs).sum(-1)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        """Check that solution is valid: nodes are not visited twice except depot and required skill levels are always met."""
        batch_size, graph_size = td["skills"].shape[0], td["skills"].shape[1]
        sorted_pi = actions.data.sort(1).values

        # Sorting it should give all zeros at front and then 1...n
        assert (
            torch.arange(1, graph_size + 1, out=sorted_pi.data.new())
            .view(1, -1)
            .expand(batch_size, graph_size)
            == sorted_pi[:, -graph_size:]
        ).all() and (sorted_pi[:, :-graph_size] == 0).all(), "Invalid tour"

        # make sure all required skill  levels are met
        indices = torch.nonzero(actions == 0)
        skills = torch.cat(
            [torch.zeros(batch_size, 1, 1, device=td.device), td["skills"]], 1
        )
        skills_ordered = gather_by_index(skills, actions).reshape(
            [batch_size, actions.size(-1), 1]
        )
        batch = start = tech = 0
        for each in indices:
            if each[0] > batch:
                start = tech = 0
                batch = each[0]
            assert (
                skills_ordered[batch, start : each[1]] <= td["techs"][batch, tech]
            ).all(), "Skill level not met"
            start = each[1] + 1  # skip the depot
            tech += 1



import os

from functools import partial
from pathlib import Path
from typing import Tuple

import torch

from tensordict import TensorDict

ProcessingData = list[Tuple[int, int]]


def list_files(path):
    import os

    files = [
        os.path.join(path, f)
        for f in os.listdir(path)
        if os.path.isfile(os.path.join(path, f))
    ]
    return files


def parse_job_line(line: Tuple[int]) -> Tuple[ProcessingData]:
    """
    Parses a FJSPLIB job data line of the following form:

        <num operations> * (<num machines> * (<machine> <processing time>))

    In words, the first value is the number of operations. Then, for each
    operation, the first number represents the number of machines that can
    process the operation, followed by, the machine index and processing time
    for each eligible machine.

    Note that the machine indices start from 1, so we subtract 1 to make them
    zero-based.
    """
    num_operations = line[0]
    operations = []
    idx = 1

    for _ in range(num_operations):
        num_pairs = int(line[idx]) * 2
        machines = line[idx + 1 : idx + 1 + num_pairs : 2]
        durations = line[idx + 2 : idx + 2 + num_pairs : 2]
        operations.append([(m, d) for m, d in zip(machines, durations)])

        idx += 1 + num_pairs

    return operations


def get_n_ops_of_instance(file):
    lines = file2lines(file)
    jobs = [parse_job_line(line) for line in lines[1:]]
    n_ope_per_job = torch.Tensor([len(x) for x in jobs]).unsqueeze(0)
    total_ops = int(n_ope_per_job.sum())
    return total_ops


def get_max_ops_from_files(files):
    return max(map(get_n_ops_of_instance, files))


def read(loc: Path, max_ops=None):
    """
    Reads an FJSPLIB instance.

    Args:
        loc: location of instance file
        max_ops: optionally specify the maximum number of total operations (will be filled by padding)

    Returns:
        instance: the parsed instance
    """
    lines = file2lines(loc)

    # First line contains metadata.
    num_jobs, num_machines = lines[0][0], lines[0][1]

    # The remaining lines contain the job-operation data, where each line
    # represents a job and its operations.
    jobs = [parse_job_line(line) for line in lines[1:]]
    n_ope_per_job = torch.Tensor([len(x) for x in jobs]).unsqueeze(0)
    total_ops = int(n_ope_per_job.sum())
    if max_ops is not None:
        assert total_ops <= max_ops, "got more operations then specified through max_ops"
    max_ops = max_ops or total_ops
    max_ops_per_job = int(n_ope_per_job.max())

    end_op_per_job = n_ope_per_job.cumsum(1) - 1
    start_op_per_job = torch.cat((torch.zeros((1, 1)), end_op_per_job[:, :-1] + 1), dim=1)

    pad_mask = torch.arange(max_ops)
    pad_mask = pad_mask.ge(total_ops).unsqueeze(0)

    proc_times = torch.zeros((num_machines, max_ops))
    op_cnt = 0
    for job in jobs:
        for op in job:
            for ma, dur in op:
                # subtract one to let indices start from zero
                proc_times[ma - 1, op_cnt] = dur
            op_cnt += 1
    proc_times = proc_times.unsqueeze(0)

    td = TensorDict(
        {
            "start_op_per_job": start_op_per_job,
            "end_op_per_job": end_op_per_job,
            "proc_times": proc_times,
            "pad_mask": pad_mask,
        },
        batch_size=[1],
    )

    return td, num_jobs, num_machines, max_ops_per_job


def file2lines(loc: Path | str) -> list[list[int]]:
    with open(loc, "r") as fh:
        lines = [line for line in fh.readlines() if line.strip()]

    def parse_num(word: str):
        return int(word) if "." not in word else int(float(word))

    return [[parse_num(x) for x in line.split()] for line in lines]


def write_one(args, where=None):
    id, instance = args
    assert (
        len(instance["proc_times"].shape) == 2
    ), "no batch dimension allowed in write operation"
    lines = []

    # The flexibility is the average number of eligible machines per operation.
    num_eligible = (instance["proc_times"] > 0).sum()
    n_ops = (~instance["pad_mask"]).sum()
    num_jobs = instance["next_op"].size(0)
    num_machines = instance["proc_times"].size(0)
    flexibility = round(int(num_eligible) / int(n_ops), 5)

    metadata = f"{num_jobs}\t{num_machines}\t{flexibility}"
    lines.append(metadata)

    for i in range(num_jobs):
        ops_of_job = instance["job_ops_adj"][i].nonzero().squeeze(1)
        job = [len(ops_of_job)]  # number of operations of the job

        for op in ops_of_job:
            eligible_ma = instance["proc_times"][:, op].nonzero().squeeze(1)
            job.append(eligible_ma.size(0))  # num_eligible

            for machine in eligible_ma:
                duration = instance["proc_times"][machine, op]
                assert duration > 0, "something is wrong"
                # add one since in song instances ma indices start from one
                job.extend([int(machine.item()) + 1, int(duration.item())])

        line = " ".join(str(num) for num in job)
        lines.append(line)

    formatted = "\n".join(lines)

    file_name = f"{str(id+1).rjust(4, '0')}_{num_jobs}j_{num_machines}m.txt"
    full_path = os.path.join(where, file_name)

    with open(full_path, "w") as fh:
        fh.write(formatted)

    return formatted


def write(where: Path | str, instances: TensorDict):
    if not os.path.exists(where):
        os.makedirs(where)

    return list(map(partial(write_one, where=where), enumerate(iter(instances))))



from typing import Callable

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


MAX_LENGTHS = {20: 2.0, 50: 3.0, 100: 4.0}


class PCTSPGenerator(Generator):
    """Data generator for the Prize-collecting Traveling Salesman Problem (PCTSP).

    Args:
        num_loc: number of locations (customers) in the VRP, without the depot. (e.g. 10 means 10 locs + 1 depot)
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        loc_distribution: distribution for the location coordinates
        depot_distribution: distribution for the depot location. If None, sample the depot from the locations
        min_demand: minimum value for the demand of each customer
        max_demand: maximum value for the demand of each customer
        demand_distribution: distribution for the demand of each customer
        capacity: capacity of the vehicle

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each city
            depot [batch_size, 2]: location of the depot
            demand [batch_size, num_loc]: demand of each customer
            capacity [batch_size, 1]: capacity of the vehicle
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        depot_distribution: int | float | str | type | Callable = None,
        penalty_factor: float = 3.0,
        prize_required: float = 1.0,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.penalty_fctor = penalty_factor
        self.prize_required = prize_required

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        # Depot distribution
        if kwargs.get("depot_sampler", None) is not None:
            self.depot_sampler = kwargs["depot_sampler"]
        else:
            self.depot_sampler = (
                get_sampler("depot", depot_distribution, min_loc, max_loc, **kwargs)
                if depot_distribution is not None
                else None
            )

        # Prize distribution
        self.deterministic_prize_sampler = get_sampler(
            "deterministric_prize", "uniform", 0.0, 4.0 / self.num_loc, **kwargs
        )
        self.stochastic_prize_sampler = get_sampler(
            "stochastic_prize", "uniform", 0.0, 2.0, **kwargs
        )

        # For the penalty to make sense it should be not too large (in which case all nodes will be visited) nor too small
        # so we want the objective term to be approximately equal to the length of the tour, which we estimate with half
        # of the nodes by half of the tour length (which is very rough but similar to op)
        # This means that the sum of penalties for all nodes will be approximately equal to the tour length (on average)
        # The expected total (uniform) penalty of half of the nodes (since approx half will be visited by the constraint)
        # is (n / 2) / 2 = n / 4 so divide by this means multiply by 4 / n,
        # However instead of 4 we use penalty_factor (3 works well) so we can make them larger or smaller
        self.max_penalty = kwargs.get("max_penalty", None)
        if self.max_penalty is None:  # If not provided, use the default max penalty
            self.max_penalty = MAX_LENGTHS.get(num_loc, None)
        if (
            self.max_penalty is None
        ):  # If not in the table keys, find the closest number of nodes as the key
            closest_num_loc = min(MAX_LENGTHS.keys(), key=lambda x: abs(x - num_loc))
            self.max_penalty = MAX_LENGTHS[closest_num_loc]
            log.warning(
                f"The max penalty for {num_loc} locations is not defined. Using the closest max penalty: {self.max_penalty}\
                    with {closest_num_loc} locations."
            )

        # Adjust as in Kool et al. (2019)
        self.max_penalty *= penalty_factor / self.num_loc
        self.penalty_sampler = get_sampler(
            "penalty", "uniform", 0.0, self.max_penalty, **kwargs
        )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations: depot and customers
        if self.depot_sampler is not None:
            depot = self.depot_sampler.sample((*batch_size, 2))
            locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))
        else:
            # if depot_sampler is None, sample the depot from the locations
            locs = self.loc_sampler.sample((*batch_size, self.num_loc + 1, 2))
            depot = locs[..., 0, :]
            locs = locs[..., 1:, :]

        # Sample penalty
        penalty = self.penalty_sampler.sample((*batch_size, self.num_loc))

        # Take uniform prizes
        # Now expectation is 0.5 so expected total prize is n / 2, we want to force to visit approximately half of the nodes
        # so the constraint will be that total prize >= (n / 2) / 2 = n / 4
        # equivalently, we divide all prizes by n / 4 and the total prize should be >= 1
        deterministic_prize = self.deterministic_prize_sampler.sample(
            (*batch_size, self.num_loc)
        )

        # In the deterministic setting, the stochastic_prize is not used and the deterministic prize is known
        # In the stochastic setting, the deterministic prize is the expected prize and is known up front but the
        # stochastic prize is only revealed once the node is visited
        # Stochastic prize is between (0, 2 * expected_prize) such that E(stochastic prize) = E(deterministic_prize)
        stochastic_prize = (
            self.stochastic_prize_sampler.sample((*batch_size, self.num_loc))
            * deterministic_prize
        )

        return TensorDict(
            {
                "locs": locs,
                "depot": depot,
                "penalty": penalty,
                "deterministic_prize": deterministic_prize,
                "stochastic_prize": stochastic_prize,
            },
            batch_size=batch_size,
        )



from typing import Iterable, Optional

import lightning.pytorch as pl
import torch

from lightning import Callback, Trainer
from lightning.fabric.accelerators.cuda import num_cuda_devices
from lightning.pytorch.accelerators import Accelerator
from lightning.pytorch.core.datamodule import LightningDataModule
from lightning.pytorch.loggers import Logger
from lightning.pytorch.strategies import DDPStrategy, Strategy
from lightning.pytorch.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS

from rl4co import utils

log = utils.get_pylogger(__name__)


class RL4COTrainer(Trainer):
    """Wrapper around Lightning Trainer, with some RL4CO magic for efficient training.

    Note:
        The most important hyperparameter to use is `reload_dataloaders_every_n_epochs`.
        This allows for datasets to be re-created on the run and distributed by Lightning across
        devices on each epoch. Setting to a value different than 1 may lead to overfitting to a
        specific (such as the initial) data distribution.

    Args:
        accelerator: hardware accelerator to use.
        callbacks: list of callbacks.
        logger: logger (or iterable collection of loggers) for experiment tracking.
        min_epochs: minimum number of training epochs.
        max_epochs: maximum number of training epochs.
        strategy: training strategy to use (if any), such as Distributed Data Parallel (DDP).
        devices: number of devices to train on (int) or which GPUs to train on (list or str) applied per node.
        gradient_clip_val: 0 means don't clip. Defaults to 1.0 for stability.
        precision: allows for mixed precision training. Can be specified as a string (e.g., '16').
            This also allows to use `FlashAttention` by default.
        disable_profiling_executor: Disable JIT profiling executor. This reduces memory and increases speed.
        auto_configure_ddp: Automatically configure DDP strategy if multiple GPUs are available.
        reload_dataloaders_every_n_epochs: Set to a value different than 1 to reload dataloaders every n epochs.
        matmul_precision: Set matmul precision for faster inference https://pytorch.org/docs/stable/generated/torch.set_float32_matmul_precision.html#torch.set_float32_matmul_precision
        **kwargs: Additional keyword arguments passed to the Lightning Trainer. See :class:`lightning.pytorch.trainer.Trainer` for details.
    """

    def __init__(
        self,
        accelerator: str | Accelerator = "auto",
        callbacks: Optional[list[Callback]] = None,
        logger: Optional[Logger | Iterable[Logger]] = None,
        min_epochs: Optional[int] = None,
        max_epochs: Optional[int] = None,
        strategy: str | Strategy = "auto",
        devices: list[int] | str | int = "auto",
        gradient_clip_val: int | float = 1.0,
        precision: str | int = "16-mixed",
        reload_dataloaders_every_n_epochs: int = 1,
        disable_profiling_executor: bool = True,
        auto_configure_ddp: bool = True,
        matmul_precision: str | int = "medium",
        **kwargs,
    ):
        # Disable JIT profiling executor. This reduces memory and increases speed.
        # Reference: https://github.com/HazyResearch/safari/blob/111d2726e7e2b8d57726b7a8b932ad8a4b2ad660/train.py#LL124-L129C17
        if disable_profiling_executor:
            try:
                torch._C._jit_set_profiling_executor(False)
                torch._C._jit_set_profiling_mode(False)
            except AttributeError:
                pass

        # Configure DDP automatically if multiple GPUs are available
        if auto_configure_ddp and strategy == "auto":
            if devices == "auto":
                n_devices = num_cuda_devices()
            elif isinstance(devices, Iterable):
                n_devices = len(devices)
            else:
                n_devices = devices
            if n_devices > 1:
                log.info(
                    "Configuring DDP strategy automatically with {} GPUs".format(
                        n_devices
                    )
                )
                strategy = DDPStrategy(
                    find_unused_parameters=True,  # We set to True due to RL envs
                    gradient_as_bucket_view=True,  # https://pytorch-lightning.readthedocs.io/en/stable/advanced/advanced_gpu.html#ddp-optimizations
                )

        # Set matmul precision for faster inference https://pytorch.org/docs/stable/generated/torch.set_float32_matmul_precision.html#torch.set_float32_matmul_precision
        if matmul_precision is not None:
            torch.set_float32_matmul_precision(matmul_precision)

        # Check if gradient_clip_val is set to None
        if gradient_clip_val is None:
            log.warning(
                "gradient_clip_val is set to None. This may lead to unstable training."
            )

        # We should reload dataloaders every epoch for RL training
        if reload_dataloaders_every_n_epochs != 1:
            log.warning(
                "We reload dataloaders every epoch for RL training. Setting reload_dataloaders_every_n_epochs to a value different than 1 "
                + "may lead to unexpected behavior since the initial conditions will be the same for `n_epochs` epochs."
            )

        # Main call to `Trainer` superclass
        super().__init__(
            accelerator=accelerator,
            callbacks=callbacks,
            logger=logger,
            min_epochs=min_epochs,
            max_epochs=max_epochs,
            strategy=strategy,
            gradient_clip_val=gradient_clip_val,
            devices=devices,
            precision=precision,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
            **kwargs,
        )

    def fit(
        self,
        model: "pl.LightningModule",
        train_dataloaders: Optional[TRAIN_DATALOADERS | LightningDataModule] = None,
        val_dataloaders: Optional[EVAL_DATALOADERS] = None,
        datamodule: Optional[LightningDataModule] = None,
        ckpt_path: Optional[str] = None,
    ) -> None:
        """
        We override the `fit` method to automatically apply and handle RL4CO magic
        to 'self.automatic_optimization = False' models, such as PPO

        It behaves exactly like the original `fit` method, but with the following changes:
        - if the given model is 'self.automatic_optimization = False', we override 'gradient_clip_val' as None
        """

        if not model.automatic_optimization:
            if self.gradient_clip_val is not None:
                log.warning(
                    "Overriding gradient_clip_val to None for 'automatic_optimization=False' models"
                )
                self.gradient_clip_val = None

        super().fit(
            model=model,
            train_dataloaders=train_dataloaders,
            val_dataloaders=val_dataloaders,
            datamodule=datamodule,
            ckpt_path=ckpt_path,
        )



from typing import Callable, Optional

import torch
import torch.nn as nn

from rl4co.models.nn.graph.attnnet import (
    MultiHeadAttentionLayer,
    Normalization,
    SkipConnection,
)
from rl4co.models.zoo.mdam.mha import MultiHeadAttentionMDAM


class MDAMGraphAttentionEncoder(nn.Module):
    def __init__(
        self,
        num_heads,
        embed_dim,
        num_layers,
        node_dim=None,
        normalization="batch",
        feedforward_hidden=512,
        sdpa_fn: Optional[Callable] = None,
    ):
        super(MDAMGraphAttentionEncoder, self).__init__()

        # To map input to embedding space
        self.init_embed = nn.Linear(node_dim, embed_dim) if node_dim is not None else None

        self.layers = nn.Sequential(
            *(
                MultiHeadAttentionLayer(
                    embed_dim,
                    num_heads,
                    feedforward_hidden,
                    normalization,
                    sdpa_fn=sdpa_fn,
                )
                for _ in range(num_layers - 1)  # because last layer is different
            )
        )
        self.attention_layer = MultiHeadAttentionMDAM(
            embed_dim, num_heads, sdpa_fn=sdpa_fn, last_one=True
        )
        self.BN1 = Normalization(embed_dim, normalization)
        self.projection = SkipConnection(
            nn.Sequential(
                nn.Linear(embed_dim, feedforward_hidden),
                nn.ReLU(),
                nn.Linear(feedforward_hidden, embed_dim),
            )
            if feedforward_hidden > 0
            else nn.Linear(embed_dim, embed_dim)
        )
        self.BN2 = Normalization(embed_dim, normalization)

    def forward(self, x, mask=None, return_transform_loss=False):
        """
        Returns:
            - h [batch_size, graph_size, embed_dim]
            - attn [num_head, batch_size, graph_size, graph_size]
            - V [num_head, batch_size, graph_size, key_dim]
            - h_old [batch_size, graph_size, embed_dim]
        """
        assert mask is None, "TODO mask not yet supported!"

        h_embeded = x
        h_old = self.layers(h_embeded)
        h_new, attn, V = self.attention_layer(h_old)
        h = h_new + h_old
        h = self.BN1(h)
        h = self.projection(h)
        h = self.BN2(h)

        return (h, h.mean(dim=1), attn, V, h_old)

    def change(self, attn, V, h_old, mask):
        num_heads, batch_size, graph_size, feat_size = V.size()
        attn = (
            mask.float()
            .view(1, batch_size, 1, graph_size)
            .repeat(num_heads, 1, graph_size, 1)
            * attn
        )
        attn = attn / (
            torch.sum(attn, dim=-1).view(num_heads, batch_size, graph_size, 1) + 1e-9
        )
        heads = torch.matmul(attn, V)

        h_new = torch.mm(
            heads.permute(1, 2, 0, 3)
            .contiguous()
            .view(-1, self.attention_layer.num_heads * self.attention_layer.val_dim),
            self.attention_layer.W_out.view(-1, self.attention_layer.embed_dim),
        ).view(batch_size, graph_size, self.attention_layer.embed_dim)
        h = h_new + h_old
        h = self.BN1(h)
        h = self.projection(h)
        h = self.BN2(h)

        return (h, h.mean(dim=1))



from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE, StepwisePPO
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline

from .policy import L2DPolicy, L2DPolicy4PPO


class L2DPPOModel(StepwisePPO):
    """Learning2Dispatch model by Zhang et al. (2020):
    'Learning to Dispatch for Job Shop Scheduling via Deep Reinforcement Learning'

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to rollout (1 epoch of exponential, then greedy rollout baseline)
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: L2DPolicy = None,
        policy_kwargs={},
        **kwargs,
    ):
        assert env.name in [
            "fjsp",
            "jssp",
        ], "L2DModel currently only works for Job-Shop Scheduling Problems"
        if policy is None:
            policy = L2DPolicy4PPO(env_name=env.name, **policy_kwargs)

        super().__init__(env, policy, **kwargs)


class L2DModel(REINFORCE):
    """Learning2Dispatch model by Zhang et al. (2020):
    'Learning to Dispatch for Job Shop Scheduling via Deep Reinforcement Learning'

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to rollout (1 epoch of exponential, then greedy rollout baseline)
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: L2DPolicy = None,
        baseline: REINFORCEBaseline | str = "rollout",
        policy_kwargs={},
        baseline_kwargs={},
        **kwargs,
    ):
        assert env.name in [
            "fjsp",
            "jssp",
        ], "L2DModel currently only works for Job-Shop Scheduling Problems"
        if policy is None:
            policy = L2DPolicy(env_name=env.name, **policy_kwargs)

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)



import os

import numpy as np

from tensordict.tensordict import TensorDict

CURR_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_PATH = os.path.dirname(os.path.dirname(CURR_DIR))


def load_npz_to_tensordict(filename):
    """Load a npz file directly into a TensorDict
    We assume that the npz file contains a dictionary of numpy arrays
    This is at least an order of magnitude faster than pickle
    """
    x = np.load(filename)
    x_dict = dict(x)
    batch_size = x_dict[list(x_dict.keys())[0]].shape[0]
    return TensorDict(x_dict, batch_size=batch_size)


def save_tensordict_to_npz(tensordict, filename, compress: bool = False):
    """Save a TensorDict to a npz file
    We assume that the TensorDict contains a dictionary of tensors
    """
    x_dict = {k: v.numpy() for k, v in tensordict.items()}
    if compress:
        np.savez_compressed(filename, **x_dict)
    else:
        np.savez(filename, **x_dict)


def check_extension(filename, extension=".npz"):
    """Check that filename has extension, otherwise add it"""
    if os.path.splitext(filename)[1] != extension:
        return filename + extension
    return filename


def load_solomon_instance(name, path=None, edge_weights=False):
    """Load solomon instance from a file"""
    import vrplib

    if not path:
        path = "data/solomon/instances/"
        path = os.path.join(ROOT_PATH, path)
    if not os.path.isdir(path):
        os.makedirs(path)
    file_path = f"{path}{name}.txt"
    if not os.path.isfile(file_path):
        vrplib.download_instance(name=name, path=path)
    return vrplib.read_instance(
        path=file_path,
        instance_format="solomon",
        compute_edge_weights=edge_weights,
    )


def load_solomon_solution(name, path=None):
    """Load solomon solution from a file"""
    import vrplib

    if not path:
        path = "data/solomon/solutions/"
        path = os.path.join(ROOT_PATH, path)
    if not os.path.isdir(path):
        os.makedirs(path)
    file_path = f"{path}{name}.sol"
    if not os.path.isfile(file_path):
        vrplib.download_solution(name=name, path=path)
    return vrplib.read_solution(path=file_path)



import random

import torch


class Cluster:
    """
    Multiple gaussian distributed clusters, as in the Solomon benchmark dataset
    Following the setting in Bi et al. 2022 (https://arxiv.org/abs/2210.07686)

    Args:
        n_cluster: Number of the gaussian distributed clusters
    """

    def __init__(self, n_cluster: int = 3):
        super().__init__()
        self.lower, self.upper = 0.2, 0.8
        self.std = 0.07
        self.n_cluster = n_cluster

    def sample(self, size):

        batch_size, num_loc, _ = size

        # Generate the centers of the clusters
        center = self.lower + (self.upper - self.lower) * torch.rand(
            batch_size, self.n_cluster * 2
        )

        # Pre-define the coordinates
        coords = torch.zeros(batch_size, num_loc, 2)

        # Calculate the size of each cluster
        cluster_sizes = [num_loc // self.n_cluster] * self.n_cluster
        for i in range(num_loc % self.n_cluster):
            cluster_sizes[i] += 1

        # Generate the coordinates
        current_index = 0
        for i in range(self.n_cluster):
            means = center[:, i * 2 : (i + 1) * 2]
            stds = torch.full((batch_size, 2), self.std)
            points = torch.normal(
                means.unsqueeze(1).expand(-1, cluster_sizes[i], -1),
                stds.unsqueeze(1).expand(-1, cluster_sizes[i], -1),
            )
            coords[:, current_index : current_index + cluster_sizes[i], :] = points
            current_index += cluster_sizes[i]

        # Confine the coordinates to range [0, 1]
        coords.clamp_(0, 1)

        return coords


class Mixed:
    """
    50% nodes sampled from uniform distribution, 50% nodes sampled from gaussian distribution, as in the Solomon benchmark dataset
    Following the setting in Bi et al. 2022 (https://arxiv.org/abs/2210.07686)

    Args:
        n_cluster_mix: Number of the gaussian distributed clusters
    """

    def __init__(self, n_cluster_mix=1):
        super().__init__()
        self.lower, self.upper = 0.2, 0.8
        self.std = 0.07
        self.n_cluster_mix = n_cluster_mix

    def sample(self, size):

        batch_size, num_loc, _ = size

        # Generate the centers of the clusters
        center = self.lower + (self.upper - self.lower) * torch.rand(
            batch_size, self.n_cluster_mix * 2
        )

        # Pre-define the coordinates sampled under uniform distribution
        coords = torch.FloatTensor(batch_size, num_loc, 2).uniform_(0, 1)

        # Sample mutated index (default setting: 50% mutation)
        mutate_idx = torch.stack(
            [torch.randperm(num_loc)[: num_loc // 2] for _ in range(batch_size)]
        )

        # Generate the coordinates
        segment_size = num_loc // (2 * self.n_cluster_mix)
        remaining_indices = num_loc // 2 - segment_size * (self.n_cluster_mix - 1)
        sizes = [segment_size] * (self.n_cluster_mix - 1) + [remaining_indices]
        for i in range(self.n_cluster_mix):
            indices = mutate_idx[:, sum(sizes[:i]) : sum(sizes[: i + 1])]
            means_x = center[:, 2 * i].unsqueeze(1).expand(-1, sizes[i])
            means_y = center[:, 2 * i + 1].unsqueeze(1).expand(-1, sizes[i])
            coords.scatter_(
                1,
                indices.unsqueeze(-1).expand(-1, -1, 2),
                torch.stack(
                    [
                        torch.normal(means_x.expand(-1, sizes[i]), self.std),
                        torch.normal(means_y.expand(-1, sizes[i]), self.std),
                    ],
                    dim=2,
                ),
            )

        # Confine the coordinates to range [0, 1]
        coords.clamp_(0, 1)

        return coords


class Gaussian_Mixture:
    """
    Following Zhou et al. (2023): https://arxiv.org/abs/2305.19587

    Args:
        num_modes: the number of clusters/modes in the Gaussian Mixture.
        cdist: scale of the uniform distribution for center generation.
    """

    def __init__(self, num_modes: int = 0, cdist: int = 0):
        super().__init__()
        self.num_modes = num_modes
        self.cdist = cdist

    def sample(self, size):

        batch_size, num_loc, _ = size

        if self.num_modes == 0:  # (0, 0) - uniform
            return torch.rand((batch_size, num_loc, 2))
        elif self.num_modes == 1 and self.cdist == 1:  # (1, 1) - gaussian
            return self.generate_gaussian(batch_size, num_loc)
        else:
            res = [self.generate_gaussian_mixture(num_loc) for _ in range(batch_size)]
            return torch.stack(res)

    def generate_gaussian_mixture(self, num_loc):
        """Following the setting in Zhang et al. 2022 (https://arxiv.org/abs/2204.03236)"""

        # Randomly decide how many points each mode gets
        nums = torch.multinomial(
            input=torch.ones(self.num_modes) / self.num_modes,
            num_samples=num_loc,
            replacement=True,
        )

        # Prepare to collect points
        coords = torch.empty((0, 2))

        # Generate points for each mode
        for i in range(self.num_modes):
            num = (nums == i).sum()  # Number of points in this mode
            if num > 0:
                center = torch.rand((1, 2)) * self.cdist
                cov = torch.eye(2)  # Covariance matrix
                nxy = torch.distributions.MultivariateNormal(
                    center.squeeze(), covariance_matrix=cov
                ).sample((num,))
                coords = torch.cat((coords, nxy), dim=0)

        return self._global_min_max_scaling(coords)

    def generate_gaussian(self, batch_size, num_loc):
        """Following the setting in Xin et al. 2022 (https://openreview.net/pdf?id=nJuzV-izmPJ)"""

        # Mean and random covariances
        mean = torch.full((batch_size, num_loc, 2), 0.5)
        covs = torch.rand(batch_size)  # Random covariances between 0 and 1

        # Generate the coordinates
        coords = torch.zeros((batch_size, num_loc, 2))
        for i in range(batch_size):
            # Construct covariance matrix for each sample
            cov_matrix = torch.tensor([[1.0, covs[i]], [covs[i], 1.0]])
            m = torch.distributions.MultivariateNormal(
                mean[i], covariance_matrix=cov_matrix
            )
            coords[i] = m.sample()

        # Shuffle the coordinates
        indices = torch.randperm(coords.size(0))
        coords = coords[indices]

        return self._batch_normalize_and_center(coords)

    def _global_min_max_scaling(self, coords):

        # Scale the points to [0, 1] using min-max scaling
        coords_min = coords.min(0, keepdim=True).values
        coords_max = coords.max(0, keepdim=True).values
        coords = (coords - coords_min) / (coords_max - coords_min)

        return coords

    def _batch_normalize_and_center(self, coords):
        # Step 1: Compute min and max along each batch
        coords_min = coords.min(dim=1, keepdim=True).values
        coords_max = coords.max(dim=1, keepdim=True).values

        # Step 2: Normalize coordinates to range [0, 1]
        coords = (
            coords - coords_min
        )  # Broadcasting subtracts min value on each coordinate
        range_max = (
            (coords_max - coords_min).max(dim=-1, keepdim=True).values
        )  # The maximum range among both coordinates
        coords = coords / range_max  # Divide by the max range to normalize

        # Step 3: Center the batch in the middle of the [0, 1] range
        coords = (
            coords + (1 - coords.max(dim=1, keepdim=True).values) / 2
        )  # Centering the batch

        return coords


class Mix_Distribution:
    """
    Mixture of three exemplar distributions in batch-level, i.e. Uniform, Cluster, Mixed
    Following the setting in Bi et al. 2022 (https://arxiv.org/abs/2210.07686)

    Args:
        n_cluster: Number of the gaussian distributed clusters in Cluster distribution
        n_cluster_mix: Number of the gaussian distributed clusters in Mixed distribution
    """

    def __init__(self, n_cluster=3, n_cluster_mix=1):
        super().__init__()
        self.lower, self.upper = 0.2, 0.8
        self.std = 0.07
        self.Mixed = Mixed(n_cluster_mix=n_cluster_mix)
        self.Cluster = Cluster(n_cluster=n_cluster)

    def sample(self, size):

        batch_size, num_loc, _ = size

        # Pre-define the coordinates sampled under uniform distribution
        coords = torch.FloatTensor(batch_size, num_loc, 2).uniform_(0, 1)

        # Random sample probability for the distribution of each sample
        p = torch.rand(batch_size)

        # Mixed
        mask = p <= 0.33
        n_mixed = mask.sum().item()
        if n_mixed > 0:
            coords[mask] = self.Mixed.sample((n_mixed, num_loc, 2))

        # Cluster
        mask = (p > 0.33) & (p <= 0.66)
        n_cluster = mask.sum().item()
        if n_cluster > 0:
            coords[mask] = self.Cluster.sample((n_cluster, num_loc, 2))

        # The remaining ones are uniformly distributed
        return coords


class Mix_Multi_Distributions:
    """
    Mixture of 11 Gaussian-like distributions in batch-level
    Following the setting in Zhou et al. (2023): https://arxiv.org/abs/2305.19587
    """

    def __init__(self):
        super().__init__()
        self.dist_set = [(0, 0), (1, 1)] + [
            (m, c) for m in [3, 5, 7] for c in [10, 30, 50]
        ]

    def sample(self, size):
        batch_size, num_loc, _ = size
        coords = torch.zeros(batch_size, num_loc, 2)

        # Pre-select distributions for the entire batch
        dists = [random.choice(self.dist_set) for _ in range(batch_size)]
        unique_dists = list(
            set(dists)
        )  # Unique distributions to minimize re-instantiation

        # Instantiate Gaussian_Mixture only once per unique distribution
        gm_instances = {dist: Gaussian_Mixture(*dist) for dist in unique_dists}

        # Batch process where possible
        for i, dist in enumerate(dists):
            coords[i] = gm_instances[dist].sample((1, num_loc, 2)).squeeze(0)

        return coords



from typing import Callable

import torch.nn as nn

from rl4co.models.common.constructive.autoregressive.policy import AutoregressivePolicy
from rl4co.models.zoo.am.decoder import AttentionModelDecoder
from rl4co.models.zoo.am.encoder import AttentionModelEncoder


class AttentionModelPolicy(AutoregressivePolicy):
    """
    Attention Model Policy based on Kool et al. (2019): https://arxiv.org/abs/1803.08475.
    This model first encodes the input graph using a Graph Attention Network (GAT) (:class:`AttentionModelEncoder`)
    and then decodes the solution using a pointer network (:class:`AttentionModelDecoder`). Cache is used to store the
    embeddings of the nodes to be used by the decoder to save computation.
    See :class:`rl4co.models.common.constructive.autoregressive.policy.AutoregressivePolicy` for more details on the inference process.

    Args:
        encoder: Encoder module, defaults to :class:`AttentionModelEncoder`
        decoder: Decoder module, defaults to :class:`AttentionModelDecoder`
        embed_dim: Dimension of the node embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads in the attention layers
        normalization: Normalization type in the attention layers
        feedforward_hidden: Dimension of the hidden layer in the feedforward network
        env_name: Name of the environment used to initialize embeddings
        encoder_network: Network to use for the encoder
        init_embedding: Module to use for the initialization of the embeddings
        context_embedding: Module to use for the context embedding
        dynamic_embedding: Module to use for the dynamic embedding
        use_graph_context: Whether to use the graph context
        linear_bias_decoder: Whether to use a bias in the linear layer of the decoder
        sdpa_fn_encoder: Function to use for the scaled dot product attention in the encoder
        sdpa_fn_decoder: Function to use for the scaled dot product attention in the decoder
        sdpa_fn: (deprecated) Function to use for the scaled dot product attention
        mask_inner: Whether to mask the inner product
        out_bias_pointer_attn: Whether to use a bias in the pointer attention
        check_nan: Whether to check for nan values during decoding
        temperature: Temperature for the softmax
        tanh_clipping: Tanh clipping value (see Bello et al., 2016)
        mask_logits: Whether to mask the logits during decoding
        train_decode_type: Type of decoding to use during training
        val_decode_type: Type of decoding to use during validation
        test_decode_type: Type of decoding to use during testing
        moe_kwargs: Keyword arguments for MoE,
            e.g., {"encoder": {"hidden_act": "ReLU", "num_experts": 4, "k": 2, "noisy_gating": True},
                   "decoder": {"light_version": True, ...}}
    """

    def __init__(
        self,
        encoder: nn.Module = None,
        decoder: nn.Module = None,
        embed_dim: int = 128,
        num_encoder_layers: int = 3,
        num_heads: int = 8,
        normalization: str = "batch",
        feedforward_hidden: int = 512,
        env_name: str = "tsp",
        encoder_network: nn.Module = None,
        init_embedding: nn.Module = None,
        context_embedding: nn.Module = None,
        dynamic_embedding: nn.Module = None,
        use_graph_context: bool = True,
        linear_bias_decoder: bool = False,
        sdpa_fn: Callable = None,
        sdpa_fn_encoder: Callable = None,
        sdpa_fn_decoder: Callable = None,
        mask_inner: bool = True,
        out_bias_pointer_attn: bool = False,
        check_nan: bool = True,
        temperature: float = 1.0,
        tanh_clipping: float = 10.0,
        mask_logits: bool = True,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "greedy",
        moe_kwargs: dict = {"encoder": None, "decoder": None},
        **unused_kwargs,
    ):
        if encoder is None:
            encoder = AttentionModelEncoder(
                embed_dim=embed_dim,
                num_heads=num_heads,
                num_layers=num_encoder_layers,
                env_name=env_name,
                normalization=normalization,
                feedforward_hidden=feedforward_hidden,
                net=encoder_network,
                init_embedding=init_embedding,
                sdpa_fn=sdpa_fn if sdpa_fn_encoder is None else sdpa_fn_encoder,
                moe_kwargs=moe_kwargs["encoder"],
            )

        if decoder is None:
            decoder = AttentionModelDecoder(
                embed_dim=embed_dim,
                num_heads=num_heads,
                env_name=env_name,
                context_embedding=context_embedding,
                dynamic_embedding=dynamic_embedding,
                sdpa_fn=sdpa_fn if sdpa_fn_decoder is None else sdpa_fn_decoder,
                mask_inner=mask_inner,
                out_bias_pointer_attn=out_bias_pointer_attn,
                linear_bias=linear_bias_decoder,
                use_graph_context=use_graph_context,
                check_nan=check_nan,
                moe_kwargs=moe_kwargs["decoder"],
            )

        super(AttentionModelPolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            temperature=temperature,
            tanh_clipping=tanh_clipping,
            mask_logits=mask_logits,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **unused_kwargs,
        )



from rl4co.models.common.constructive.base import ConstructivePolicy

from .decoder import AutoregressiveDecoder
from .encoder import AutoregressiveEncoder


class AutoregressivePolicy(ConstructivePolicy):
    """Template class for an autoregressive policy, simple wrapper around
    :class:`rl4co.models.common.constructive.base.ConstructivePolicy`.

    Note:
        While a decoder is required, an encoder is optional and will be initialized to
        :class:`rl4co.models.common.constructive.autoregressive.encoder.NoEncoder`.
        This can be used in decoder-only models in which at each step actions do not depend on
        previously encoded states.
    """

    def __init__(
        self,
        encoder: AutoregressiveEncoder,
        decoder: AutoregressiveDecoder,
        env_name: str = "tsp",
        temperature: float = 1.0,
        tanh_clipping: float = 0,
        mask_logits: bool = True,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "greedy",
        **unused_kw,
    ):
        # We raise an error for the user if no decoder was provided
        if decoder is None:
            raise ValueError("AutoregressivePolicy requires a decoder to be provided.")

        super(AutoregressivePolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            temperature=temperature,
            tanh_clipping=tanh_clipping,
            mask_logits=mask_logits,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **unused_kw,
        )



import math

import torch
import torch.nn as nn

from tensordict import TensorDict

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.models.common.improvement.base import ImprovementPolicy
from rl4co.models.zoo.n2s.encoder import N2SEncoder
from rl4co.models.zoo.neuopt.decoder import RDSDecoder
from rl4co.utils.decoding import DecodingStrategy, get_decoding_strategy
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class CustomizeTSPInitEmbedding(nn.Module):
    """Initial embedding for the Traveling Salesman Problems (TSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the cities
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(CustomizeTSPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Sequential(
            nn.Linear(node_dim, embed_dim // 2, linear_bias),
            nn.ReLU(inplace=True),
            nn.Linear(embed_dim // 2, embed_dim, linear_bias),
        )

    def forward(self, td):
        out = self.init_embed(td["locs"])
        return out


class NeuOptPolicy(ImprovementPolicy):
    """
    NeuOpt Policy based on Ma et al. (2023)
    This model first encodes the input graph and current solution using a N2S encoder (:class:`N2SEncoder`)
    and then decodes the k-opt action (:class:`RDSDecoder`)

    Args:
        embed_dim: Dimension of the node embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads in the attention layers
        normalization: Normalization type in the attention layers
        feedforward_hidden: Dimension of the hidden layer in the feedforward network
        env_name: Name of the environment used to initialize embeddings
        pos_type: Name of the used positional encoding method (CPE or APE)
        init_embedding: Module to use for the initialization of the embeddings
        pos_embedding: Module to use for the initialization of the positional embeddings
        temperature: Temperature for the softmax
        tanh_clipping: Tanh clipping value (see Bello et al., 2016)
        train_decode_type: Type of decoding to use during training
        val_decode_type: Type of decoding to use during validation
        test_decode_type: Type of decoding to use during testing
    """

    def __init__(
        self,
        embed_dim: int = 128,
        num_encoder_layers: int = 3,
        num_heads: int = 4,
        normalization: str = "layer",
        feedforward_hidden: int = 128,
        env_name: str = "tsp_kopt",
        pos_type: str = "CPE",
        init_embedding: nn.Module = None,
        pos_embedding: nn.Module = None,
        temperature: float = 1.0,
        tanh_clipping: float = 6.0,
        train_decode_type: str = "sampling",
        val_decode_type: str = "sampling",
        test_decode_type: str = "sampling",
    ):
        super(NeuOptPolicy, self).__init__()

        self.env_name = env_name
        self.embed_dim = embed_dim

        # Decoding strategies
        self.temperature = temperature
        self.tanh_clipping = tanh_clipping
        self.train_decode_type = train_decode_type
        self.val_decode_type = val_decode_type
        self.test_decode_type = test_decode_type

        # Encoder and decoder
        if init_embedding is None:
            init_embedding = CustomizeTSPInitEmbedding(self.embed_dim)

        self.encoder = N2SEncoder(
            embed_dim=embed_dim,
            init_embedding=init_embedding,
            pos_embedding=pos_embedding,
            env_name=env_name,
            pos_type=pos_type,
            num_heads=num_heads,
            num_layers=num_encoder_layers,
            normalization=normalization,
            feedforward_hidden=feedforward_hidden,
        )

        self.decoder = RDSDecoder(embed_dim=embed_dim)

        self.init_hidden_W = nn.Linear(self.embed_dim, self.embed_dim)
        self.init_query_learnable = nn.Parameter(torch.Tensor(self.embed_dim))

        self.init_parameters()

    def init_parameters(self) -> None:
        for param in self.parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(
        self,
        td: TensorDict,
        env: str | RL4COEnvBase = None,
        phase: str = "train",
        return_actions: bool = True,
        return_embeds: bool = False,
        only_return_embed: bool = False,
        actions=None,
        **decoding_kwargs,
    ) -> dict:
        """Forward pass of the policy.

        Args:
            td: TensorDict containing the environment state
            env: Environment to use for decoding. If None, the environment is instantiated from `env_name`. Note that
                it is more efficient to pass an already instantiated environment each time for fine-grained control
            phase: Phase of the algorithm (train, val, test)
            return_actions: Whether to return the actions
            actions: Actions to use for evaluating the policy.
                If passed, use these actions instead of sampling from the policy to calculate log likelihood
            decoding_kwargs: Keyword arguments for the decoding strategy. See :class:`rl4co.utils.decoding.DecodingStrategy` for more information.

        Returns:
            out: Dictionary containing the reward, log likelihood, and optionally the actions and entropy
        """

        # Encoder: get encoder output and initial embeddings from initial state
        nfe, _ = self.encoder(td)
        if only_return_embed:
            return {"embeds": nfe.detach()}

        # Instantiate environment if needed
        if isinstance(env, str) or env is None:
            env_name = self.env_name if env is None else env
            log.info(f"Instantiated environment not provided; instantiating {env_name}")
            env = get_env(env_name)
        assert not env.two_opt_mode, "NeuOpt only support k-opt with k > 2"

        # Get decode type depending on phase and whether actions are passed for evaluation
        decode_type = decoding_kwargs.pop("decode_type", None)
        if actions is not None:
            decode_type = "evaluate"
        elif decode_type is None:
            decode_type = getattr(self, f"{phase}_decode_type")

        # Setup decoding strategy
        # we pop arguments that are not part of the decoding strategy
        decode_strategy: DecodingStrategy = get_decoding_strategy(
            decode_type,
            temperature=decoding_kwargs.pop("temperature", self.temperature),
            tanh_clipping=decoding_kwargs.pop("tanh_clipping", self.tanh_clipping),
            mask_logits=True,
            improvement_method_mode=True,
            **decoding_kwargs,
        )

        # Perform the decoding
        bs, gs, _, ll, action_sampled, rec, visited_time = (
            *nfe.size(),
            0.0,
            None,
            td["rec_current"],
            td["visited_time"],
        )
        action_index = torch.zeros(bs, env.k_max, dtype=torch.long).to(rec.device)
        k_action_left = torch.zeros(bs, env.k_max + 1, dtype=torch.long).to(rec.device)
        k_action_right = torch.zeros(bs, env.k_max, dtype=torch.long).to(rec.device)
        next_of_last_action = (
            torch.zeros_like(rec[:, :1], dtype=torch.long).to(rec.device) - 1
        )
        mask = torch.zeros_like(rec, dtype=torch.bool).to(rec.device)
        stopped = torch.ones(bs, dtype=torch.bool).to(rec.device)
        zeros = torch.zeros((bs, 1), device=td.device)

        # init queries
        h_mean = nfe.mean(1)
        init_query = self.init_query_learnable.repeat(bs, 1)
        input_q1 = input_q2 = init_query.clone()
        init_hidden = self.init_hidden_W(h_mean)
        q1 = q2 = init_hidden.clone()

        for i in range(env.k_max):
            # Pass RDS decoder
            logits, q1, q2 = self.decoder(nfe, q1, q2, input_q1, input_q2)

            # Calc probs
            if i == 0 and "action" in td.keys():
                mask = mask.scatter(1, td["action"][:, :1], 1)

            logprob, action_sampled = decode_strategy.step(
                logits,
                ~mask.clone(),
                action=actions[:, i : i + 1].squeeze() if actions is not None else None,
            )
            action_sampled = action_sampled.unsqueeze(-1)
            if i > 0:
                action_sampled = torch.where(
                    stopped.unsqueeze(-1), action_index[:, :1], action_sampled
                )
            if phase == "train":
                loss_now = logprob.gather(1, action_sampled)
            else:
                loss_now = zeros.clone()

            # Record log_likelihood and Entropy
            if i > 0:
                ll = ll + torch.where(stopped.unsqueeze(-1), zeros * 0, loss_now)
            else:
                ll = ll + loss_now

            # Store and Process actions
            next_of_new_action = rec.gather(1, action_sampled)
            action_index[:, i] = action_sampled.squeeze().clone()
            k_action_left[stopped, i] = action_sampled[stopped].squeeze().clone()
            k_action_right[~stopped, i - 1] = action_sampled[~stopped].squeeze().clone()
            k_action_left[:, i + 1] = next_of_new_action.squeeze().clone()

            # Prepare next RNN input
            input_q1 = nfe.gather(
                1, action_sampled.view(bs, 1, 1).expand(bs, 1, self.embed_dim)
            ).squeeze(1)
            input_q2 = torch.where(
                stopped.view(bs, 1).expand(bs, self.embed_dim),
                input_q1.clone(),
                nfe.gather(
                    1,
                    (next_of_last_action % gs)
                    .view(bs, 1, 1)
                    .expand(bs, 1, self.embed_dim),
                ).squeeze(1),
            )

            # Process if k-opt close
            # assert (input_q1[stopped] == input_q2[stopped]).all()
            if i > 0:
                stopped = stopped | (action_sampled == next_of_last_action).squeeze()
            else:
                stopped = (action_sampled == next_of_last_action).squeeze()
            # assert (input_q1[stopped] == input_q2[stopped]).all()

            k_action_left[stopped, i] = k_action_left[stopped, i - 1]
            k_action_right[stopped, i] = k_action_right[stopped, i - 1]

            # Calc next basic masks
            if i == 0:
                visited_time_tag = (
                    visited_time - visited_time.gather(1, action_sampled)
                ) % gs
            mask &= False
            mask[(visited_time_tag <= visited_time_tag.gather(1, action_sampled))] = True
            if i == 0:
                mask[visited_time_tag > (gs - 2)] = True
            mask[stopped, action_sampled[stopped].squeeze()] = (
                False  # allow next k-opt starts immediately
            )
            # if True:#i == env.k_max - 2: # allow special case: close k-opt at the first selected node
            index_allow_first_node = (~stopped) & (
                next_of_new_action.squeeze() == action_index[:, 0]
            )
            mask[index_allow_first_node, action_index[index_allow_first_node, 0]] = False

            # Move to next
            next_of_last_action = next_of_new_action
            next_of_last_action[stopped] = -1

        # Form final action
        k_action_right[~stopped, -1] = k_action_left[~stopped, -1].clone()
        k_action_left = k_action_left[:, : env.k_max]
        action_all = torch.cat((action_index, k_action_left, k_action_right), -1)

        outdict = {"log_likelihood": ll, "cost_bsf": td["cost_bsf"]}
        td.set("action", action_all)

        if return_embeds:
            outdict["embeds"] = nfe.detach()

        if return_actions:
            outdict["actions"] = action_all

        return outdict



import itertools
import math
import warnings

from typing import Callable, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange

from rl4co.models.nn.moe import MoE
from rl4co.utils import get_pylogger

log = get_pylogger(__name__)


def scaled_dot_product_attention_simple(
    q, k, v, attn_mask=None, dropout_p=0.0, is_causal=False
):
    """Simple (exact) Scaled Dot-Product Attention in RL4CO without customized kernels (i.e. no Flash Attention)."""

    # Check for causal and attn_mask conflict
    if is_causal and attn_mask is not None:
        raise ValueError("Cannot set both is_causal and attn_mask")

    # Calculate scaled dot product
    scores = torch.matmul(q, k.transpose(-2, -1)) / (k.size(-1) ** 0.5)

    # Apply the provided attention mask
    if attn_mask is not None:
        if attn_mask.dtype == torch.bool:
            scores.masked_fill_(~attn_mask, float("-inf"))
        else:
            scores += attn_mask

    # Apply causal mask
    if is_causal:
        s, l_ = scores.size(-2), scores.size(-1)
        mask = torch.triu(torch.ones((s, l_), device=scores.device), diagonal=1)
        scores.masked_fill_(mask.bool(), float("-inf"))

    # Softmax to get attention weights
    attn_weights = F.softmax(scores, dim=-1)

    # Apply dropout
    if dropout_p > 0.0:
        attn_weights = F.dropout(attn_weights, p=dropout_p)

    # Compute the weighted sum of values
    return torch.matmul(attn_weights, v)


try:
    from torch.nn.functional import scaled_dot_product_attention
except ImportError:
    log.warning(
        "torch.nn.functional.scaled_dot_product_attention not found. Make sure you are using PyTorch >= 2.0.0."
        "Alternatively, install Flash Attention https://github.com/HazyResearch/flash-attention ."
        "Using custom implementation of scaled_dot_product_attention without Flash Attention. "
    )
    scaled_dot_product_attention = scaled_dot_product_attention_simple


class MultiHeadAttention(nn.Module):
    """PyTorch native implementation of Flash Multi-Head Attention with automatic mixed precision support.
    Uses PyTorch's native `scaled_dot_product_attention` implementation, available from 2.0

    Note:
        If `scaled_dot_product_attention` is not available, use custom implementation of `scaled_dot_product_attention` without Flash Attention.

    Args:
        embed_dim: total dimension of the model
        num_heads: number of heads
        bias: whether to use bias
        attention_dropout: dropout rate for attention weights
        causal: whether to apply causal mask to attention scores
        device: torch device
        dtype: torch dtype
        sdpa_fn: scaled dot product attention function (SDPA) implementation
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        bias: bool = True,
        attention_dropout: float = 0.0,
        causal: bool = False,
        device: str = None,
        dtype: torch.dtype = None,
        sdpa_fn: Optional[Callable] = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.causal = causal
        self.attention_dropout = attention_dropout
        self.sdpa_fn = sdpa_fn if sdpa_fn is not None else scaled_dot_product_attention

        self.num_heads = num_heads
        assert self.embed_dim % num_heads == 0, "self.kdim must be divisible by num_heads"
        self.head_dim = self.embed_dim // num_heads
        assert (
            self.head_dim % 8 == 0 and self.head_dim <= 128
        ), "Only support head_dim <= 128 and divisible by 8"

        self.Wqkv = nn.Linear(embed_dim, 3 * embed_dim, bias=bias, **factory_kwargs)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias, **factory_kwargs)

    def forward(self, x, attn_mask=None):
        """x: (batch, seqlen, hidden_dim) (where hidden_dim = num heads * head dim)
        attn_mask: bool tensor of shape (batch, seqlen)
        """
        # Project query, key, value
        q, k, v = rearrange(
            self.Wqkv(x), "b s (three h d) -> three b h s d", three=3, h=self.num_heads
        ).unbind(dim=0)

        if attn_mask is not None:
            attn_mask = (
                attn_mask.unsqueeze(1)
                if attn_mask.ndim == 3
                else attn_mask.unsqueeze(1).unsqueeze(2)
            )

        # Scaled dot product attention
        out = self.sdpa_fn(
            q,
            k,
            v,
            attn_mask=attn_mask,
            dropout_p=self.attention_dropout,
        )
        return self.out_proj(rearrange(out, "b h s d -> b s (h d)"))


def sdpa_fn_wrapper(q, k, v, attn_mask=None, dmat=None, dropout_p=0.0, is_causal=False):
    if dmat is not None:
        log.warning(
            "Edge weights passed to simple attention-fn, which is not supported. Weights will be ignored..."
        )
    return scaled_dot_product_attention(
        q, k, v, attn_mask=attn_mask, dropout_p=dropout_p, is_causal=is_causal
    )


class MultiHeadCrossAttention(nn.Module):
    """PyTorch native implementation of Flash Multi-Head Cross Attention with automatic mixed precision support.
    Uses PyTorch's native `scaled_dot_product_attention` implementation, available from 2.0

    Note:
        If `scaled_dot_product_attention` is not available, use custom implementation of `scaled_dot_product_attention` without Flash Attention.

    Args:
        embed_dim: total dimension of the model
        num_heads: number of heads
        bias: whether to use bias
        attention_dropout: dropout rate for attention weights
        device: torch device
        dtype: torch dtype
        sdpa_fn: scaled dot product attention function (SDPA)
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        bias: bool = False,
        attention_dropout: float = 0.0,
        device: str = None,
        dtype: torch.dtype = None,
        sdpa_fn: Optional[Callable | nn.Module] = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.attention_dropout = attention_dropout

        # Default to `scaled_dot_product_attention` if `sdpa_fn` is not provided
        if sdpa_fn is None:
            sdpa_fn = sdpa_fn_wrapper
        self.sdpa_fn = sdpa_fn

        self.num_heads = num_heads
        assert self.embed_dim % num_heads == 0, "self.kdim must be divisible by num_heads"
        self.head_dim = self.embed_dim // num_heads
        assert (
            self.head_dim % 8 == 0 and self.head_dim <= 128
        ), "Only support head_dim <= 128 and divisible by 8"

        self.Wq = nn.Linear(embed_dim, embed_dim, bias=bias, **factory_kwargs)
        self.Wkv = nn.Linear(embed_dim, 2 * embed_dim, bias=bias, **factory_kwargs)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias, **factory_kwargs)

    def forward(self, q_input, kv_input, cross_attn_mask=None, dmat=None):
        # Project query, key, value
        q = rearrange(
            self.Wq(q_input), "b m (h d) -> b h m d", h=self.num_heads
        )  # [b, h, m, d]
        k, v = rearrange(
            self.Wkv(kv_input), "b n (two h d) -> two b h n d", two=2, h=self.num_heads
        ).unbind(
            dim=0
        )  # [b, h, n, d]

        if cross_attn_mask is not None:
            # add head dim
            cross_attn_mask = cross_attn_mask.unsqueeze(1)

        # Scaled dot product attention
        out = self.sdpa_fn(
            q,
            k,
            v,
            attn_mask=cross_attn_mask,
            dmat=dmat,
            dropout_p=self.attention_dropout,
        )
        return self.out_proj(rearrange(out, "b h s d -> b s (h d)"))


class PointerAttention(nn.Module):
    """Calculate logits given query, key and value and logit key.
    This follows the pointer mechanism of Vinyals et al. (2015) (https://arxiv.org/abs/1506.03134).

    Note:
        With Flash Attention, masking is not supported

    Performs the following:
        1. Apply cross attention to get the heads
        2. Project heads to get glimpse
        3. Compute attention score between glimpse and logit key

    Args:
        embed_dim: total dimension of the model
        num_heads: number of heads
        mask_inner: whether to mask inner attention
        linear_bias: whether to use bias in linear projection
        check_nan: whether to check for NaNs in logits
        sdpa_fn: scaled dot product attention function (SDPA) implementation
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        mask_inner: bool = True,
        out_bias: bool = False,
        check_nan: bool = True,
        sdpa_fn: Callable | str = "default",
        **kwargs,
    ):
        super(PointerAttention, self).__init__()
        self.num_heads = num_heads
        self.mask_inner = mask_inner

        # Projection - query, key, value already include projections
        self.project_out = nn.Linear(embed_dim, embed_dim, bias=out_bias)
        self.check_nan = check_nan

        # Defaults for sdpa_fn implementation
        # see https://github.com/ai4co/rl4co/issues/228
        if isinstance(sdpa_fn, str):
            if sdpa_fn == "default":
                sdpa_fn = scaled_dot_product_attention
            elif sdpa_fn == "simple":
                sdpa_fn = scaled_dot_product_attention_simple
            else:
                raise ValueError(
                    f"Unknown sdpa_fn: {sdpa_fn}. Available options: ['default', 'simple']"
                )
        else:
            if sdpa_fn is None:
                sdpa_fn = scaled_dot_product_attention
                log.info(
                    "Using default scaled_dot_product_attention for PointerAttention"
                )
        self.sdpa_fn = sdpa_fn

    def forward(self, query, key, value, logit_key, attn_mask=None):
        """Compute attention logits given query, key, value, logit key and attention mask.

        Args:
            query: query tensor of shape [B, ..., L, E]
            key: key tensor of shape [B, ..., S, E]
            value: value tensor of shape [B, ..., S, E]
            logit_key: logit key tensor of shape [B, ..., S, E]
            attn_mask: attention mask tensor of shape [B, ..., S]. Note that `True` means that the value _should_ take part in attention
                as described in the [PyTorch Documentation](https://pytorch.org/docs/stable/generated/torch.nn.functional.scaled_dot_product_attention.html)
        """
        # Compute inner multi-head attention with no projections.
        heads = self._inner_mha(query, key, value, attn_mask)
        glimpse = self._project_out(heads, attn_mask)

        # Batch matrix multiplication to compute logits (batch_size, num_steps, graph_size)
        # bmm is slightly faster than einsum and matmul
        logits = (torch.bmm(glimpse, logit_key.squeeze(-2).transpose(-2, -1))).squeeze(
            -2
        ) / math.sqrt(glimpse.size(-1))

        if self.check_nan:
            assert not torch.isnan(logits).any(), "Logits contain NaNs"

        return logits

    def _inner_mha(self, query, key, value, attn_mask):
        q = self._make_heads(query)
        k = self._make_heads(key)
        v = self._make_heads(value)
        if self.mask_inner:
            # make mask the same number of dimensions as q
            attn_mask = (
                attn_mask.unsqueeze(1)
                if attn_mask.ndim == 3
                else attn_mask.unsqueeze(1).unsqueeze(2)
            )
        else:
            attn_mask = None
        heads = self.sdpa_fn(q, k, v, attn_mask=attn_mask)
        return rearrange(heads, "... h n g -> ... n (h g)", h=self.num_heads)

    def _make_heads(self, v):
        return rearrange(v, "... g (h s) -> ... h g s", h=self.num_heads)

    def _project_out(self, out, *kwargs):
        return self.project_out(out)


class PointerAttnMoE(PointerAttention):
    """Calculate logits given query, key and value and logit key.
    This follows the pointer mechanism of Vinyals et al. (2015) <https://arxiv.org/abs/1506.03134>,
        and the MoE gating mechanism of Zhou et al. (2024) <https://arxiv.org/abs/2405.01029>.

    Note:
        With Flash Attention, masking is not supported

    Performs the following:
        1. Apply cross attention to get the heads
        2. Project heads to get glimpse
        3. Compute attention score between glimpse and logit key

    Args:
        embed_dim: total dimension of the model
        num_heads: number of heads
        mask_inner: whether to mask inner attention
        linear_bias: whether to use bias in linear projection
        check_nan: whether to check for NaNs in logits
        sdpa_fn: scaled dot product attention function (SDPA) implementation
        moe_kwargs: Keyword arguments for MoE
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        mask_inner: bool = True,
        out_bias: bool = False,
        check_nan: bool = True,
        sdpa_fn: Optional[Callable] = None,
        moe_kwargs: Optional[dict] = None,
    ):
        super(PointerAttnMoE, self).__init__(
            embed_dim, num_heads, mask_inner, out_bias, check_nan, sdpa_fn
        )
        self.moe_kwargs = moe_kwargs

        self.project_out = None
        self.project_out_moe = MoE(
            embed_dim, embed_dim, num_neurons=[], out_bias=out_bias, **moe_kwargs
        )
        if self.moe_kwargs["light_version"]:
            self.dense_or_moe = nn.Linear(embed_dim, 2, bias=False)
            self.project_out = nn.Linear(embed_dim, embed_dim, bias=out_bias)

    def _project_out(self, out, attn_mask):
        """Implementation of Hierarchical Gating based on Zhou et al. (2024) <https://arxiv.org/abs/2405.01029>."""
        if self.moe_kwargs["light_version"]:
            num_nodes, num_available_nodes = attn_mask.size(-1), attn_mask.sum(-1)
            # only do this at the "second" step, which is depot -> pomo -> first select
            if (num_available_nodes >= num_nodes - 1).any():
                self.probs = F.softmax(
                    self.dense_or_moe(
                        out.view(-1, out.size(-1)).mean(dim=0, keepdim=True)
                    ),
                    dim=-1,
                )
            selected = self.probs.multinomial(1).squeeze(0)
            out = (
                self.project_out_moe(out)
                if selected.item() == 1
                else self.project_out(out)
            )
            glimpse = out * self.probs.squeeze(0)[selected]
        else:
            glimpse = self.project_out_moe(out)
        return glimpse


# Deprecated
class LogitAttention(PointerAttention):
    def __init__(self, *args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn(
            "LogitAttention is deprecated and will be removed in a future release. "
            "Please use PointerAttention instead."
            "Note that several components of the previous LogitAttention have moved to `rl4co.models.nn.dec_strategies`.",
            category=DeprecationWarning,
        )
        super(LogitAttention, self).__init__(*args, **kwargs)


# MultiHeadCompat
class MultiHeadCompat(nn.Module):
    def __init__(self, n_heads, input_dim, embed_dim=None, val_dim=None, key_dim=None):
        super(MultiHeadCompat, self).__init__()

        if val_dim is None:
            # assert embed_dim is not None, "Provide either embed_dim or val_dim"
            val_dim = embed_dim // n_heads
        if key_dim is None:
            key_dim = val_dim

        self.n_heads = n_heads
        self.input_dim = input_dim
        self.embed_dim = embed_dim
        self.val_dim = val_dim
        self.key_dim = key_dim

        self.W_query = nn.Parameter(torch.Tensor(n_heads, input_dim, key_dim))
        self.W_key = nn.Parameter(torch.Tensor(n_heads, input_dim, key_dim))

        self.init_parameters()

    # used for init nn.Parameter
    def init_parameters(self):
        for param in self.parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(self, q, h=None, mask=None):
        """

        :param q: queries (batch_size, n_query, input_dim)
        :param h: data (batch_size, graph_size, input_dim)
        :param mask: mask (batch_size, n_query, graph_size) or viewable as that (i.e. can be 2 dim if n_query == 1)
        Mask should contain 1 if attention is not possible (i.e. mask is negative adjacency)
        :return:
        """

        if h is None:
            h = q  # compute self-attention

        # h should be (batch_size, graph_size, input_dim)
        batch_size, graph_size, input_dim = h.size()
        n_query = q.size(1)

        hflat = h.contiguous().view(-1, input_dim)  #################   reshape
        qflat = q.contiguous().view(-1, input_dim)

        # last dimension can be different for keys and values
        shp = (self.n_heads, batch_size, graph_size, -1)
        shp_q = (self.n_heads, batch_size, n_query, -1)

        # Calculate queries, (n_heads, n_query, graph_size, key/val_size)
        Q = torch.matmul(qflat, self.W_query).view(shp_q)
        K = torch.matmul(hflat, self.W_key).view(shp)

        # Calculate compatibility (n_heads, batch_size, n_query, graph_size)
        compatibility_s2n = torch.matmul(Q, K.transpose(2, 3))

        return compatibility_s2n


class PolyNetAttention(PointerAttention):
    """Calculate logits given query, key and value and logit key.
    This implements a modified version the pointer mechanism of Vinyals et al. (2015) (https://arxiv.org/abs/1506.03134)
    as described in Hottung et al. (2024) (https://arxiv.org/abs/2402.14048) PolyNetAttention conditions the attention logits on
    a set of k different binary vectors allowing to learn k different solution strategies.

    Note:
        With Flash Attention, masking is not supported

    Performs the following:
        1. Apply cross attention to get the heads
        2. Project heads to get glimpse
        3. Apply PolyNet layers
        4. Compute attention score between glimpse and logit key

    Args:
        k: Number unique bit vectors used to compute attention score
        embed_dim: total dimension of the model
        poly_layer_dim: Dimension of the PolyNet layers
        num_heads: number of heads
        mask_inner: whether to mask inner attention
        linear_bias: whether to use bias in linear projection
        check_nan: whether to check for NaNs in logits
        sdpa_fn: scaled dot product attention function (SDPA) implementation
    """

    def __init__(
        self, k: int, embed_dim: int, poly_layer_dim: int, num_heads: int, **kwargs
    ):
        super(PolyNetAttention, self).__init__(embed_dim, num_heads, **kwargs)

        self.k = k
        self.binary_vector_dim = math.ceil(math.log2(k))
        self.binary_vectors = torch.nn.Parameter(
            torch.Tensor(
                list(itertools.product([0, 1], repeat=self.binary_vector_dim))[:k]
            ),
            requires_grad=False,
        )

        self.poly_layer_1 = nn.Linear(embed_dim + self.binary_vector_dim, poly_layer_dim)
        self.poly_layer_2 = nn.Linear(poly_layer_dim, embed_dim)

    def forward(self, query, key, value, logit_key, attn_mask=None):
        """Compute attention logits given query, key, value, logit key and attention mask.

        Args:
            query: query tensor of shape [B, ..., L, E]
            key: key tensor of shape [B, ..., S, E]
            value: value tensor of shape [B, ..., S, E]
            logit_key: logit key tensor of shape [B, ..., S, E]
            attn_mask: attention mask tensor of shape [B, ..., S]. Note that `True` means that the value _should_ take part in attention
                as described in the [PyTorch Documentation](https://pytorch.org/docs/stable/generated/torch.nn.functional.scaled_dot_product_attention.html)
        """
        # Compute inner multi-head attention with no projections.
        heads = self._inner_mha(query, key, value, attn_mask)
        glimpse = self.project_out(heads)

        num_solutions = glimpse.shape[1]
        z = self.binary_vectors.repeat(math.ceil(num_solutions / self.k), 1)[
            :num_solutions
        ]
        z = z[None].expand(glimpse.shape[0], num_solutions, self.binary_vector_dim)

        # PolyNet layers
        poly_out = self.poly_layer_1(torch.cat((glimpse, z), dim=2))
        poly_out = F.relu(poly_out)
        poly_out = self.poly_layer_2(poly_out)

        glimpse += poly_out

        # Batch matrix multiplication to compute logits (batch_size, num_steps, graph_size)
        # bmm is slightly faster than einsum and matmul
        logits = (torch.bmm(glimpse, logit_key.squeeze(-2).transpose(-2, -1))).squeeze(
            -2
        ) / math.sqrt(glimpse.size(-1))

        if self.check_nan:
            assert not torch.isnan(logits).any(), "Logits contain NaNs"

        return logits



from typing import Any, Optional, Union

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.glop.policy import GLOPPolicy
from rl4co.utils.ops import gather_by_index, unbatchify


class GLOP(REINFORCE):
    """Implements GLOP: https://arxiv.org/abs/2312.08224

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to mean
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: Optional[GLOPPolicy] = None,
        baseline: Union[REINFORCEBaseline, str] = "mean",
        policy_kwargs={},
        baseline_kwargs={},
        **kwargs,
    ):
        if policy is None:
            policy = GLOPPolicy(env_name=env.name, **policy_kwargs)

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: Optional[int] = None
    ):
        td = self.env.reset(batch)
        n_samples = self.policy.n_samples

        # Evaluate policy
        out = self.policy(
            td=td,
            env=self.env,
            phase=phase,
            return_actions=True,
        )
        reward = out["reward"] = unbatchify(out["reward"], n_samples)
        max_reward, max_idxs = reward.max(dim=-1)
        out.update({"max_reward": max_reward})

        if phase == "train":
            assert n_samples > 1, "num_starts must be > 1 during training"
            log_likelihood = unbatchify(out["log_likelihood"], n_samples)
            advantage = reward - reward.mean(-1, keepdim=True)
            out["loss"] = -(advantage * log_likelihood).mean()
        else:
            if n_samples > 1 and out.get("actions", None) is not None:
                actions = unbatchify(out["actions"], n_samples)
                out.update(
                    {
                        "best_multistart_actions": gather_by_index(
                            actions, max_idxs, dim=max_idxs.dim()
                        )
                    }
                )
                out["actions"] = actions

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}



from rl4co.models.zoo.glop.adapter.tsp_adapter import TSP2SHPPAdapter

try:
    from rl4co.models.zoo.glop.adapter.vrp_adapter import VRP2SubTSPAdapter
except ImportError:
    # In case some dependencies are not installed (e.g., numba)
    VRP2SubTSPAdapter = None


adapter_map = {
    "cvrp": VRP2SubTSPAdapter,
    "cvrpmvc": VRP2SubTSPAdapter,
    "tsp": TSP2SHPPAdapter,
}



from typing import Optional, Type, Union

from tensordict import TensorDict
import torch

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.models.zoo.deepaco import DeepACOPolicy
from rl4co.models.zoo.deepaco.antsystem import AntSystem
from rl4co.models.zoo.gfacs.encoder import GFACSEncoder
from rl4co.utils.decoding import (
    DecodingStrategy,
    get_decoding_strategy,
    get_log_likelihood,
    modify_logits_for_top_k_filtering,
    modify_logits_for_top_p_filtering
)
from rl4co.utils.ops import batchify, unbatchify
from rl4co.utils.pylogger import get_pylogger


log = get_pylogger(__name__)


class GFACSPolicy(DeepACOPolicy):
    """Implememts GFACS policy based on :class:`NonAutoregressivePolicy`. Introduced by Kim et al. (2024): https://arxiv.org/abs/2403.07041.
    This policy uses a Non-Autoregressive Graph Neural Network to generate heatmaps,
    which are then used to run Ant Colony Optimization (ACO) to construct solutions.

    Args:
        encoder: Encoder module. Can be passed by sub-classes
        env_name: Name of the environment used to initialize embeddings
        temperature: Temperature for the softmax during decoding. Defaults to 1.0.
        aco_class: Class representing the ACO algorithm to be used. Defaults to :class:`AntSystem`.
        aco_kwargs: Additional arguments to be passed to the ACO algorithm.
        train_with_local_search: Whether to train with local search. Defaults to False.
        n_ants: Number of ants to be used in the ACO algorithm. Can be an integer or dictionary. Defaults to 20.
        n_iterations: Number of iterations to run the ACO algorithm. Can be an integer or dictionary. Defaults to `dict(train=1, val=20, test=100)`.
        encoder_kwargs: Additional arguments to be passed to the encoder.
    """

    def __init__(
        self,
        encoder: Optional[GFACSEncoder] = None,
        env_name: str = "tsp",
        temperature: float = 1.0,
        top_p: float = 0.0,
        top_k: int = 0,
        aco_class: Optional[Type[AntSystem]] = None,
        aco_kwargs: dict = {},
        train_with_local_search: bool = True,
        n_ants: Optional[Union[int, dict]] = None,
        n_iterations: Optional[Union[int, dict]] = None,
        multistart: bool = False,
        k_sparse: Optional[int] = None,
        **encoder_kwargs,
    ):
        if encoder is None:
            encoder_kwargs["z_out_dim"] = 2 if train_with_local_search else 1
            encoder_kwargs["k_sparse"] = k_sparse
            encoder = GFACSEncoder(env_name=env_name, **encoder_kwargs)

        super().__init__(
            encoder=encoder,
            env_name=env_name,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            aco_class=aco_class,
            aco_kwargs=aco_kwargs,
            train_with_local_search=train_with_local_search,
            n_ants=n_ants,
            n_iterations=n_iterations,
            multistart=multistart,
            k_sparse=k_sparse,
        )

    def forward(
        self,
        td_initial: TensorDict,
        env: Optional[Union[str, RL4COEnvBase]] = None,
        phase: str = "train",
        return_actions: bool = True,
        return_hidden: bool = False,
        actions=None,
        **decoding_kwargs,
    ) -> dict:
        """
        Forward method. During validation and testing, the policy runs the ACO algorithm to construct solutions.
        See :class:`NonAutoregressivePolicy` for more details during the training phase.
        """
        n_ants = self.n_ants[phase]

        heatmap, _, logZ = self.encoder(td_initial)

        decoding_kwargs.update(self.default_decoding_kwargs)
        decoding_kwargs.update(
            {"num_starts": n_ants} if "multistart" in self.decode_type else {"num_samples": n_ants}
        )

        # Instantiate environment if needed
        if (phase != "train" or self.train_with_local_search) and (env is None or isinstance(env, str)):
            env_name = self.env_name if env is None else env
            env = get_env(env_name)
        else:
            assert isinstance(env, RL4COEnvBase), "env must be an instance of RL4COEnvBase"

        if phase == "train":
            # Encoder: get encoder output and initial embeddings from initial state
            if self.train_with_local_search:
                logZ, ls_logZ = logZ[:, [0]], logZ[:, [1]]
            else:
                logZ = logZ[:, [0]]

            logprobs, actions, td, env = self.common_decoding(
                self.decode_type, td_initial, env, heatmap, actions, **decoding_kwargs
            )

            # Output dictionary construction
            outdict = {
                "logZ": logZ,
                "reward": unbatchify(env.get_reward(td, actions), n_ants),
                "log_likelihood": unbatchify(
                    get_log_likelihood(logprobs, actions, td.get("mask", None), True), n_ants
                )
            }

            if return_actions:
                outdict["actions"] = actions

            ########################################################################
            # Local search
            if self.train_with_local_search:
                # TODO: Refactor this so that we don't need to use the aco object
                aco = self.aco_class(heatmap, n_ants=n_ants, **self.aco_kwargs)
                ls_actions, ls_reward = aco.local_search(
                    batchify(td_initial, n_ants), env, actions, decoding_kwargs
                )
                ls_decoding_kwargs = decoding_kwargs.copy()
                ls_decoding_kwargs["top_k"] = 0  # This should be 0, otherwise logprobs can be -inf
                ls_logprobs, ls_actions, td, env = self.common_decoding(
                    "evaluate", td_initial, env, heatmap, ls_actions, **ls_decoding_kwargs
                )
                outdict.update(
                    {
                        "ls_logZ": ls_logZ,
                        "ls_reward": unbatchify(ls_reward, n_ants),
                        "ls_log_likelihood": unbatchify(
                            get_log_likelihood(ls_logprobs, ls_actions, td.get("mask", None), True),
                            n_ants,
                        )
                    }
                )
                if return_actions:
                    outdict["ls_actions"] = ls_actions
            ########################################################################

            if return_hidden:
                outdict["hidden"] = heatmap

            return outdict

        heatmap /= self.temperature

        if self.top_k > 0:
            self.top_k = min(self.top_k, heatmap.size(-1))  # safety check
            heatmap = modify_logits_for_top_k_filtering(heatmap, self.top_k)

        if self.top_p > 0:
            assert self.top_p <= 1.0, "top-p should be in (0, 1]."
            heatmap = modify_logits_for_top_p_filtering(heatmap, self.top_p)

        aco = self.aco_class(heatmap, n_ants=n_ants, **self.aco_kwargs)
        actions, iter_rewards = aco.run(td_initial, env, self.n_iterations[phase], decoding_kwargs)

        out = {"reward": iter_rewards[self.n_iterations[phase] - 1]}
        out.update({f"reward_{i:03d}": iter_rewards[i] for i in range(self.n_iterations[phase])})
        if return_actions:
            out["actions"] = actions

        return out

    def common_decoding(
        self,
        decode_type: str | DecodingStrategy,
        td: TensorDict,
        env: RL4COEnvBase,
        hidden: TensorDict,
        actions: Optional[torch.Tensor] = None,
        max_steps: int = 1_000_000,
        **decoding_kwargs,
    ):
        decoding_strategy: DecodingStrategy = get_decoding_strategy(
            decoding_strategy=decode_type,
            temperature=decoding_kwargs.pop("temperature", self.temperature),
            mask_logits=decoding_kwargs.pop("mask_logits", self.mask_logits),
            tanh_clipping=decoding_kwargs.pop("tanh_clipping", self.tanh_clipping),
            select_start_nodes_fn=decoding_kwargs.pop("select_start_nodes_fn", None),
            store_all_logp=decoding_kwargs.pop("store_all_logp", False),
            **decoding_kwargs,
        )
        if actions is not None:
            assert decoding_strategy.name == "evaluate", "decoding strategy must be 'evaluate' when actions are provided"

        # Pre-decoding hook: used for the initial step(s) of the decoding strategy
        td, env, num_starts = decoding_strategy.pre_decoder_hook(
            td, env, actions[:, 0] if actions is not None and "multistart" in self.decode_type else None
        )

        # Additionally call a decoder hook if needed before main decoding
        td, env, hidden = self.decoder.pre_decoder_hook(td, env, hidden, num_starts)

        # Main decoding: loop until all sequences are done
        step = 1 if "multistart" in self.decode_type else 0
        while not td["done"].all():
            logits, mask = self.decoder(td, hidden, num_starts)
            td = decoding_strategy.step(
                logits,
                mask,
                td,
                action=actions[..., step] if actions is not None else None,
            )
            td = env.step(td)["next"]
            step += 1
            if step > max_steps:
                log.error(
                    f"Exceeded maximum number of steps ({max_steps}) duing decoding"
                )
                break

        # Post-decoding hook: used for the final step(s) of the decoding strategy
        logprobs, actions, td, env = decoding_strategy.post_decoder_hook(td, env)
        return logprobs, actions, td, env


from typing import Tuple

import torch
import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor
from torch_geometric.data import Batch, Data
from torch_geometric.nn import MessagePassing

from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.mlp import MLP
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MessagePassingLayer(MessagePassing):
    def __init__(
        self,
        node_indim,
        node_outdim,
        edge_indim,
        edge_outdim,
        aggregation="add",
        residual=False,
        **mlp_params,
    ):
        super(MessagePassingLayer, self).__init__(aggr=aggregation)
        # Init message passing models
        self.edge_model = MLP(
            input_dim=edge_indim + 2 * node_indim, output_dim=edge_outdim, **mlp_params
        )
        self.node_model = MLP(
            input_dim=edge_outdim + node_indim, output_dim=node_outdim, **mlp_params
        )
        self.residual = residual

    def forward(self, node_feature, edge_feature, edge_index, mask=None):
        # Message passing
        update_edge_feature = self.edge_update(node_feature, edge_feature, edge_index)
        update_node_feature = self.propagate(
            edge_index, x=node_feature, edge_features=update_edge_feature
        )

        # Update with residual connection
        if self.residual:
            update_node_feature = update_node_feature + node_feature

        return update_node_feature, update_edge_feature

    def edge_update(self, nf, ef, edge_index):
        row, col = edge_index
        x_i, x_j = nf[row], nf[col]
        uef = self.edge_model(torch.cat([x_i, x_j, ef], dim=-1))
        return uef

    def message(self, edge_features: torch.tensor):
        return edge_features

    def update(self, aggr_msg: torch.tensor, x: torch.tensor):
        unf = self.node_model(torch.cat([x, aggr_msg], dim=-1))
        return unf


class MessagePassingEncoder(nn.Module):
    def __init__(
        self,
        env_name: str,
        embed_dim: int,
        num_nodes: int,
        num_layers: int,
        init_embedding: nn.Module = None,
        aggregation: str = "add",
        self_loop: bool = False,
        residual: bool = True,
    ):
        """
        Note:
            - Support fully connected graph for now.
        """
        super(MessagePassingEncoder, self).__init__()

        self.env_name = env_name

        self.init_embedding = (
            env_init_embedding(self.env_name, {"embed_dim": embed_dim})
            if init_embedding is None
            else init_embedding
        )

        # Generate edge index for a fully connected graph
        adj_matrix = torch.ones(num_nodes, num_nodes)
        if self_loop:
            adj_matrix.fill_diagonal_(0)  # No self-loops
        self.edge_index = torch.permute(torch.nonzero(adj_matrix), (1, 0))

        # Init message passing models
        self.mpnn_layers = nn.ModuleList(
            [
                MessagePassingLayer(
                    node_indim=embed_dim,
                    node_outdim=embed_dim,
                    edge_indim=1,
                    edge_outdim=1,
                    aggregation=aggregation,
                    residual=residual,
                )
                for _ in range(num_layers)
            ]
        )

        # Record parameters
        self.self_loop = self_loop

    # def forward(self, x, mask=None):
    def forward(
        self, td: TensorDict, mask: Tensor | None = None
    ) -> Tuple[Tensor, Tensor]:
        init_h = self.init_embedding(td)
        num_node = init_h.size(-2)

        # Check to update the edge index with different number of node
        if num_node != self.edge_index.max().item() + 1:
            adj_matrix = torch.ones(num_node, num_node)
            if self.self_loop:
                adj_matrix.fill_diagonal_(0)
            edge_index = torch.permute(torch.nonzero(adj_matrix), (1, 0))
            edge_index = edge_index.to(init_h.device)
        else:
            edge_index = self.edge_index.to(init_h.device)

        # Generate edge features: distance
        edge_feature = torch.norm(
            init_h[..., edge_index[0], :] - init_h[..., edge_index[1], :],
            dim=-1,
            keepdim=True,
        )

        # Create the batched graph
        data_list = [
            Data(x=x, edge_index=edge_index, edge_attr=edge_attr)
            for x, edge_attr in zip(init_h, edge_feature)
        ]
        data_batch = Batch.from_data_list(data_list)
        update_node_feature = data_batch.x
        update_edge_feature = data_batch.edge_attr
        edge_index = data_batch.edge_index

        # Message passing
        for layer in self.mpnn_layers:
            update_node_feature, update_edge_feature = layer(
                update_node_feature, update_edge_feature, edge_index
            )

        # De-batch the graph
        input_size = init_h.size()
        update_node_feature = update_node_feature.view(*input_size)

        return update_node_feature, init_h

    def edge_update(self, nf, ef, edge_index):
        row, col = edge_index
        x_i, x_j = nf[row], nf[col]
        uef = self.edge_model(torch.cat([x_i, x_j, ef], dim=-1))
        return uef

    def message(self, edge_features: torch.tensor):
        return edge_features

    def update(self, aggr_msg: torch.tensor, x: torch.tensor):
        unf = self.node_model(torch.cat([x, aggr_msg], dim=-1))
        return unf



from rl4co.envs.graph.flp.env import FLPEnv
from rl4co.envs.graph.flp.generator import FLPGenerator
from rl4co.envs.graph.mcp.env import MCPEnv
from rl4co.envs.graph.mcp.generator import MCPGenerator



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index

from .generator import MDCPDPGenerator
from .render import render


class MDCPDPEnv(RL4COEnvBase):
    """Multi Depot Capacitated Pickup and Delivery Problem (MDCPDP) environment.
    One reference to understand the problem could be: Solving the multi-compartment capacitated location routing
    problem with pickup–delivery routes and stochastic demands (https://doi.org/10.1016/j.cie.2015.05.008).
    This problem is also solved in a parallel manner (multi-agent) in PARCO: https://github.com/ai4co/parco. This
    environment is a single-agent version of the environment in PARCO environment.

    The environment is made of num_loc + num_agents locations (cities):
        - num_agents depot
        - num_loc / 2 pickup locations
        - num_loc / 2 delivery locations
    The goal is to visit all the pickup and delivery locations in the shortest path possible starting from the depot
    The conditions is that the agent must visit a pickup location before visiting its corresponding delivery location
    The capacity is the maximum number of pickups that the vehicle can carry at the same time

    Observations:
        - locs: locations of the cities [num_loc + num_agents, 2]
        - current_node: current node of the agent [1]
        - to_deliver: if the node is to deliver [1]
        - i: current step [1]
        - action_mask: mask of the available actions [num_loc + num_agents]
        - shape: shape of the observation

    Constraints:
        - The agent cannot visit the same city twice
        - The agent must visit the pickup location before the delivery location
        - The agent must visit the depot at the end of the tour

    Finish Condition:
        - The agent visited all the locations

    Reward:
        - Min-sum: the reward is the negative of the length of the tour
        - Min-max: the reward is the negative of the maximum length of the tour
        - Lateness: the reward is the negative of the cumulate sum of the length of the tour
        - Lateness-square: the reward is the negative of the cumulate sum of the square of the length of the tour

    Args:
        generator: MDCPDPGenerator instance as the data generator
        generator_params: parameters for the generator
        dist_mode: distance mode. One of ["L1", "L2"]
        reward_mode: objective of the problem. One of ["lateness", "lateness_square", "minmax", "minsum"]
        problem_mode: type of the problem. One of ["close", "open"]
        start_mode: type of the start. One of ["order", "random"]
    """

    name = "mdcpdp"

    def __init__(
        self,
        generator: MDCPDPGenerator = None,
        generator_params: dict = {},
        dist_mode: str = "L2",
        reward_mode: str = "lateness",
        problem_mode: str = "open",
        start_mode: str = "order",  # TODO: actually now it could be chosen by model
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = MDCPDPGenerator(**generator_params)
        self.generator = generator
        self.dist_mode = dist_mode
        self.reward_mode = reward_mode
        self.problem_mode = problem_mode
        self.start_mode = start_mode
        self.depot_mode = generator.depot_mode
        self._make_spec(self.generator)

        assert self.dist_mode in ["L1", "L2"], "Distance mode (L1/L2) not supported"
        assert self.reward_mode in [
            "lateness",
            "lateness_square",
            "minmax",
            "minsum",
        ], "Objective mode not supported"
        assert self.problem_mode in [
            "close",
            "open",
        ], "Task type (open/close) not supported"
        assert self.start_mode in [
            "order",
            "random",
        ], "Start type (order/random) not supported"

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"].unsqueeze(-1)
        current_depot = td["current_depot"]

        num_agents = td["capacity"].shape[-1]
        num_loc = td["locs"].shape[-2] - num_agents  # no depot
        pd_split_idx = num_loc // 2 + num_agents

        # Pickup and delivery node pair of selected node
        new_to_deliver = (current_node + num_loc // 2) % (num_loc + num_agents)

        back_flag = (
            (current_node < num_agents) & (current_node == current_depot) & (td["i"] > 0)
        )  # only for later steps

        # Set available to 0 (i.e., we visited the node)
        # do this only if td["i"] > 0, i.e., not the first step
        # TODO: better way?
        available = td["available"]
        if td["i"][0] > 0:
            available = available.scatter(
                -1, current_node.expand_as(td["action_mask"]), 0
            )

        # Record the to be delivered node
        to_deliver = td["to_deliver"].scatter(
            -1, new_to_deliver.expand_as(td["to_deliver"]), 1
        )

        # Update number of current carry orders
        current_carry = td["current_carry"]
        current_carry += (
            (current_node < pd_split_idx) & (current_node >= num_agents)
        ).long()  # If pickup, add 1
        current_carry -= (current_node >= pd_split_idx).long()  # If delivery, minus 1

        # Update the current depot
        # current_depot = td["current_depot"]
        # current_depot = torch.where(back_flag, current_node, current_depot)
        current_depot = torch.where(
            current_node < num_agents, current_node, td["current_depot"]
        )

        # Update the length of current tour
        current_length = td["current_length"]
        prev_loc = gather_by_index(td["locs"], td["current_node"])
        curr_loc = gather_by_index(td["locs"], current_node)

        # TODO: actually, in case of `close`, this may not always hold
        # please check if this is correct
        current_step_length = self.get_distance(prev_loc, curr_loc)[..., None]
        # If this path is the way between two depots, i.e. open a new route, set the length to 0
        current_step_length = torch.where(
            (current_node < num_agents) & (td["current_node"] < num_agents),
            0,
            current_step_length,
        )

        # If the problem mode is open, the path back to the depot will not be counted
        if self.problem_mode == "open":
            current_step_length = torch.where(
                (current_node < num_agents) & (td["current_node"] >= num_agents),
                0,
                current_step_length,
            )

        # Update the current length
        current_length.scatter_add_(-1, current_depot, current_step_length)

        # Update the arrive time for each city
        arrivetime_record = td["arrivetime_record"]
        arrivetime_record.scatter_(
            -1, current_node, current_length.gather(-1, current_depot)
        )

        # Action is feasible if the node is not visited and is to deliver
        action_mask = available & to_deliver

        # If reach the capacity, only delivery is available
        current_capacity = td["capacity"].gather(-1, current_depot)
        over_capacity = current_carry >= current_capacity
        action_mask[
            ..., num_agents:pd_split_idx
        ] &= ~over_capacity  # If reach the capacity, pickup is not available

        # If back to the current depot, this tour is done, set other depots to availbe to start
        # a new tour. Must start from a depot.
        action_mask[..., num_agents:] &= ~back_flag.expand_as(
            action_mask[..., num_agents:]
        )

        # If back to the depot, other unvisited depots are available
        # if not back to the depot, depots are not available except the current depot
        action_mask[..., :num_agents] &= back_flag.expand_as(
            action_mask[..., :num_agents]
        )
        action_mask[..., :num_agents].scatter_(-1, current_depot, ~back_flag)

        # If this is the last agent, it has to finish all the left taks
        last_depot_flag = (
            torch.sum(available[..., :num_agents].long(), dim=-1, keepdim=True) == 0
        )
        action_mask[..., :num_agents] &= ~last_depot_flag.expand_as(
            action_mask[..., :num_agents]
        )

        # Update depot mask
        carry_flag = current_carry > 0  # If agent is carrying orders
        action_mask[
            ..., :num_agents
        ] &= ~carry_flag  # If carrying orders, depot is not available

        # 1) current node is a depot
        # 2) we did not just come back
        # 3) it is not the first step
        # cannot go to other depots
        prev_depot_flag = (current_node < num_agents) & (td["i"] > 0) & ~back_flag
        action_mask[..., :num_agents] &= ~prev_depot_flag.expand_as(
            action_mask[..., :num_agents]
        )

        # We are done there are no unvisited locations
        # done = torch.count_nonzero(available, dim=-1) == 0
        done = torch.sum(available[..., num_agents:], dim=-1) == 0

        # If done, the last depot would be always available
        action_mask[..., :num_agents].scatter_(
            -1,
            current_depot,
            action_mask[..., :num_agents].gather(-1, current_depot)
            | done[..., None].expand_as(action_mask[..., :num_agents]),
        )

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        # Update step
        td.update(
            {
                "current_node": current_node,
                "current_depot": current_depot,
                "current_carry": current_carry,
                "available": available,
                "to_deliver": to_deliver,
                "i": td["i"] + 1,
                "action_mask": action_mask,
                "reward": reward,
                "done": done,
            }
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        if "depots" in td:
            depots = td["depots"]
        else:
            depots = td["depot"]

        locs = torch.cat((depots, td["locs"]), -2)

        num_agents = depots.shape[-2]
        num_loc = td["locs"].shape[-2]

        # Record how many depots are visited
        depot_idx = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        # Pick is 1, deliver is 0 [batch_size, graph_size+1], i.e. [1, 1, ..., 1, 0, ...0]
        to_deliver = torch.cat(
            [
                torch.ones(
                    *batch_size,
                    num_loc // 2 + num_agents,
                    dtype=torch.bool,
                    device=device,
                ),
                torch.zeros(
                    *batch_size,
                    num_loc // 2,
                    dtype=torch.bool,
                    device=device,
                ),
            ],
            dim=-1,
        )

        # Current depot index
        if self.start_mode == "random":
            current_depot = torch.randint(
                low=0, high=num_agents, size=(*batch_size, 1), device=device
            )
        elif self.start_mode == "order":
            current_depot = torch.zeros(
                (*batch_size, 1), dtype=torch.int64, device=device
            )

        # Current carry order number
        current_carry = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        # Current length of each depot
        current_length = torch.zeros(
            (*batch_size, num_agents), dtype=torch.float32, device=device
        )

        # Arrive time for each city
        arrivetime_record = torch.zeros(
            (*batch_size, num_loc + num_agents),
            dtype=torch.float32,
            device=device,
        )

        # Cannot visit depot at first step # [0,1...1] so set not available
        available = torch.ones(
            (*batch_size, num_loc + num_agents),
            dtype=torch.bool,
            device=device,
        )
        action_mask = ~available.contiguous()  # [batch_size, graph_size+1]
        # action_mask[..., 0] = 1  # First step is always the depot

        # Allow the model to choose any depot at the beginning
        action_mask[..., :num_agents] = True

        # Other variables
        current_node = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)
        i = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "locs": locs,
                "depot_idx": depot_idx,
                "current_node": current_node,
                "current_depot": current_depot,
                "current_carry": current_carry,
                "current_length": current_length,
                "arrivetime_record": arrivetime_record,
                "capacity": td["capacity"],
                "lateness_weight": td["lateness_weight"],
                "to_deliver": to_deliver,
                "available": available,
                "i": i,
                "action_mask": action_mask,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: MDCPDPGenerator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            to_deliver=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc + 1),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc + 1,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def get_distance(self, prev_loc, cur_loc):
        # Use L1 norm to calculate the distance for Manhattan distance
        if self.dist_mode == "L1":
            return torch.abs(cur_loc - prev_loc).norm(p=1, dim=-1)
        elif self.dist_mode == "L2":
            return torch.abs(cur_loc - prev_loc).norm(p=2, dim=-1)
        else:
            raise ValueError(f"Invalid distance norm: {self.dist_norm}")

    def _get_reward(self, td: TensorDict, actions) -> TensorDict:
        """Return the rewrad for the current state
        Support modes:
            - minmax: the reward is the maximum length of all agents
            - minsum: the reward is the sum of all agents' length
            - lateness: the reward is the sum of all agents' length plus the lateness with a weight
        Args:
            - actions [batch_size, num_agents+num_locs-1]: the actions taken by the agents
                note that the last city back to depot is not included here
        """
        # Check the validity of the actions
        num_agents = td["capacity"].shape[-1]
        num_loc = td["locs"].shape[-2] - num_agents  # except depot

        # Append the last depot to the end of the actions
        actions = torch.cat([actions, td["current_depot"]], dim=-1)

        # Calculate the reward
        if self.reward_mode == "minmax":
            cost = torch.max(td["current_length"], dim=-1)[0]
        elif self.reward_mode == "minsum":
            cost = torch.sum(td["current_length"], dim=-1)
        elif self.reward_mode == "lateness":
            cost = torch.sum(td["current_length"], dim=(-1))
            lateness = td["arrivetime_record"][..., num_agents + num_loc // 2 :]
            if self.reward_mode == "lateness_square":
                lateness = lateness**2
            lateness = torch.sum(lateness, dim=-1)
            # lateness weight - note that if this is 0, the reward is the same as the cost
            # and if this is 1, the reward is the same as the lateness
            cost = (
                cost * (1 - td["lateness_weight"].squeeze())
                + lateness * td["lateness_weight"].squeeze()
            )
        else:
            raise NotImplementedError(
                f"Invalid reward mode: {self.reward_mode}. Available modes: minmax, minsum, lateness_square, lateness"
            )
        return -cost  # minus for reward

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        assert True, "Not implemented"

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)



from torch.utils.data import DataLoader

from rl4co.envs import (
    CVRPEnv,
    CVRPTWEnv,
    DPPEnv,
    FLPEnv,
    MCPEnv,
    MDPPEnv,
    MTSPEnv,
    OPEnv,
    PCTSPEnv,
    PDPEnv,
    PDPRuinRepairEnv,
    SDVRPEnv,
    SMTWTPEnv,
    SPCTSPEnv,
    TSPEnv,
)


def get_env(name, size):
    match name:
        case "tsp":
            env = TSPEnv(generator_params=dict(num_loc=size))
        case "cvrp":
            env = CVRPEnv(generator_params=dict(num_loc=size))
        case "cvrptw":
            env = CVRPTWEnv(generator_params=dict(num_loc=size))
        case "sdvrp":
            env = SDVRPEnv(generator_params=dict(num_loc=size))
        case "pdp":
            env = PDPEnv(generator_params=dict(num_loc=size))
        case "op":
            env = OPEnv(generator_params=dict(num_loc=size))
        case "mtsp":
            env = MTSPEnv(generator_params=dict(num_loc=size))
        case "pctsp":
            env = PCTSPEnv(generator_params=dict(num_loc=size))
        case "spctsp":
            env = SPCTSPEnv(generator_params=dict(num_loc=size))
        case "dpp":
            env = DPPEnv()
        case "mdpp":
            env = MDPPEnv()
        case "smtwtp":
            env = SMTWTPEnv()
        case "pdp_ruin_repair":
            env = PDPRuinRepairEnv()
        case "mcp":
            env = MCPEnv()
        case "flp":
            env = FLPEnv()
        case _:
            raise ValueError(f"Unknown env_name: {name}")

    return env.transform()


def generate_env_data(env, size, batch_size):
    env = get_env(env, size)
    dataset = env.dataset([batch_size])

    dataloader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=0,
        collate_fn=dataset.collate_fn,
    )

    return env, next(iter(dataloader))



from functools import partial
from multiprocessing import Pool

from tensordict.tensordict import TensorDict
from torch import Tensor

from .utils import process_instance


class NoSolver:
    def solve(self, *args, **kwargs):
        pass


try:
    import rl4co.envs.routing.mtvrp.baselines.pyvrp as pyvrp
except ImportError:
    pyvrp = NoSolver()
try:
    import rl4co.envs.routing.mtvrp.baselines.lkh as lkh
except ImportError:
    lkh = NoSolver()
try:
    import rl4co.envs.routing.mtvrp.baselines.ortools as ortools
except ImportError:
    ortools = NoSolver()


def solve(
    instances: TensorDict,
    max_runtime: float,
    num_procs: int = 1,
    solver: str = "pyvrp",
    **kwargs,
) -> tuple[Tensor, Tensor]:
    """
    Solves the AnyVRP instances with PyVRP.

    Args:
        instances: The AnyVRP instances to solve.
        max_runtime: The maximum runtime for the solver.
        num_procs: The number of processes to use.
        solver: The solver to use.

    Returns:
        A tuple containing the action and the cost, respectively.
    """

    instances = process_instance(instances)

    if solver == "pyvrp" and isinstance(pyvrp, NoSolver):
        raise ImportError(
            "PyVRP is not installed. Please install it using `pip install -e .[solvers]`."
        )
    if solver == "lkh" and isinstance(lkh, NoSolver):
        raise ImportError(
            "LKH is not installed. Please install it using `pip install -e .[solvers]`"
        )
    if solver == "ortools" and isinstance(ortools, NoSolver):
        raise ImportError(
            "OR-Tools is not installed. Please install it using `pip install -e .[solvers]`."
        )

    solvers = {"pyvrp": pyvrp.solve, "ortools": ortools.solve, "lkh": lkh.solve}
    if solver not in solvers:
        raise ValueError(f"Unknown baseline solver: {solver}")

    _solve = solvers[solver]
    func = partial(_solve, max_runtime=max_runtime, **kwargs)

    if num_procs > 1:
        with Pool(processes=num_procs) as pool:
            results = pool.map(func, instances)
    else:
        results = [func(instance) for instance in instances]

    actions, costs = zip(*results)

    # Pad to ensure all actions have the same length.
    max_len = max(len(action) for action in actions)
    actions = [action + [0] * (max_len - len(action)) for action in actions]

    return Tensor(actions).long(), Tensor(costs)



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np

from matplotlib.colors import ListedColormap
from tensordict.tensordict import TensorDict

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td: TensorDict, idx: int):
    inst = td[idx]
    num_jobs = inst["job_ops_adj"].size(0)

    # Define a colormap with a color for each job
    colors = plt.cm.tab10(np.linspace(0, 1, num_jobs))
    cmap = ListedColormap(colors)

    assign = inst["ma_assignment"].nonzero()

    schedule = defaultdict(list)

    for val in assign:
        machine = val[0].item()
        op = val[1].item()
        # get start and end times of operation
        start = inst["start_times"][val[1]]
        end = inst["finish_times"][val[1]]
        # write information to schedule dictionary
        schedule[machine].append((op, start, end))

    _, ax = plt.subplots()

    # Plot horizontal bars for each task
    for ma, ops in schedule.items():
        for op, start, end in ops:
            job = inst["job_ops_adj"][:, op].nonzero().item()
            ax.barh(
                ma,
                end - start,
                left=start,
                height=0.6,
                color=cmap(job),
                edgecolor="black",
                linewidth=1,
            )

            ax.text(
                start + (end - start) / 2, ma, op, ha="center", va="center", color="white"
            )

    # Set labels and title
    ax.set_yticks(range(len(schedule)))
    ax.set_yticklabels([f"Machine {i}" for i in range(len(schedule))])
    ax.set_xlabel("Time")
    ax.set_title("Gantt Chart")

    # Add a legend for class labels
    handles = [plt.Rectangle((0, 0), 1, 1, color=cmap(i)) for i in range(num_jobs)]
    ax.legend(
        handles,
        [f"Job {label}" for label in range(num_jobs)],
        loc="center left",
        bbox_to_anchor=(1, 0.5),
    )

    plt.tight_layout()
    return ax
